﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using static CharaEdit40.Define;

namespace CharaEdit40
{
    public partial class DlgParts : Form
    {
        const int charaBaseX = 8;
        const int charaBaseY = 8;
        const int charaSize = 16;

        private int numColor = 7;
        private Pen penCursor = new Pen(Color.Red, 2);
        private Timer timerCursor = new Timer();

        private Chara chr;
        private Point pos;
        private bool front;
        private Bitmap pc80chara = new Bitmap(256, 256);
        private bool reEdnterTextBox = false;

        // -------------------------------------------------------------
        // コンストラクタ
        // c デフォルト文字コード, p 画面上の位置
        // -------------------------------------------------------------
        public DlgParts(Chara c, Point p, bool f)
        {
            InitializeComponent();
            chr = c;
            pos = p;
            front = f;

            // 文字画像を予め引き延ばしておく
            using (Pen penGrid = new Pen(Color.FromArgb(130, 130, 130)))
            using (Graphics g = Graphics.FromImage(pc80chara))
            {
                // キャラ文字画像を拡大描画する
                g.InterpolationMode = InterpolationMode.NearestNeighbor;

                Rectangle rcDest = new Rectangle(0, 1, 257, 257);
                g.DrawImage(Properties.Resources._80font, rcDest, 0, 0, 128, 128, GraphicsUnit.Pixel);

                // グリッドを描画する
                g.DrawLine(Pens.Black, charaSize, 0, 256, 0);
                for (int y = charaSize; y < 256; y += charaSize)
                {
                    g.DrawLine(penGrid, 0, y, 256, y);
                }
                for (int x = charaSize; x < 256; x += charaSize)
                {
                    g.DrawLine(penGrid, x, 0, x, 256);
                }
            }
        }

        // -------------------------------------------------------------
        // フォームがロードされる
        // -------------------------------------------------------------
        private void DlgParts_Load(object sender, EventArgs e)
        {
            textBox.Location = new Point(-100000, -100000);
            textBox.Focus();

            ClientSize = new Size(256 + charaBaseX * 2, 256 + charaBaseY * 2 + 22);
            labelHelp.Location = new Point(8, 256 + charaBaseY * 2 + 2);

            pos.X -= (chr.code % 16) * charaSize + charaBaseX;
            pos.Y -= (chr.code / 16) * charaSize + charaBaseY;
            Location = pos;

            // カーソル点滅
            timerCursor.Tick += new EventHandler(OnTickCursor);
            timerCursor.Interval = 50;
            timerCursor.Start();
        }

        // -------------------------------------------------------------
        // 何かキーが押された（フック）
        // -------------------------------------------------------------
        [System.Security.Permissions.UIPermission(
            System.Security.Permissions.SecurityAction.Demand,
            Window = System.Security.Permissions.UIPermissionWindow.AllWindows)]
        protected override bool ProcessDialogKey(Keys keyData)
        {
            // 直接コマンド
            bool keyCtrl = (ModifierKeys & Keys.Control) == Keys.Control;
            bool keyAlt = (ModifierKeys & Keys.Alt) == Keys.Alt;
            if (!keyAlt) return false;
            switch (keyData & Keys.KeyCode)
            {
                case Keys.D0:       // [0] 黒
                case Keys.NumPad0:
                    chr.color = ExColor.Black;
                    Invalidate();
                    return true;
                case Keys.D1:       // [1] 青
                case Keys.NumPad1:
                    chr.color = ExColor.Blue;
                    Invalidate();
                    return true;
                case Keys.D2:       // [2] 赤
                case Keys.NumPad2:
                    chr.color = ExColor.Red;
                    Invalidate();
                    return true;
                case Keys.D3:       // [3] 紫
                case Keys.NumPad3:
                    chr.color = ExColor.Magenta;
                    Invalidate();
                    return true;
                case Keys.D4:       // [4] 緑
                case Keys.NumPad4:
                    chr.color = ExColor.Green;
                    Invalidate();
                    return true;
                case Keys.D5:       // [5] 水色
                case Keys.NumPad5:
                    chr.color = ExColor.Cyan;
                    Invalidate();
                    return true;
                case Keys.D6:       // [6] 黄
                case Keys.NumPad6:
                    chr.color = ExColor.Yellow;
                    Invalidate();
                    return true;
                case Keys.D7:       // [7] 白
                case Keys.NumPad7:
                    chr.color = ExColor.White;
                    Invalidate();
                    return true;
            }
            return false;
        }

        // -------------------------------------------------------------
        // カーソル点滅タイマー
        // -------------------------------------------------------------
        private void OnTickCursor(object sender, EventArgs e)
        {
            numColor = (numColor % 7) + 1;
            penCursor = new Pen(MainForm.GetColor((ExColor)numColor), 2);
            Invalidate();
        }

        // -------------------------------------------------------------
        // 選択されている文字コードを取得する
        // -------------------------------------------------------------
        private int GetCharaCode(Point pnt)
        {
            pnt.X -= charaBaseX;
            pnt.Y -= charaBaseY;
            if (pnt.X < 0 || pnt.Y < 0) return -100;
            pnt.X /= charaSize;
            pnt.Y /= charaSize;
            if (pnt.X >= 16 || pnt.Y >= 16) return -100;
            return pnt.X + pnt.Y * 16;
        }

        // -------------------------------------------------------------
        // 文字コードから選択範囲を取得する
        // -------------------------------------------------------------
        private Rectangle GetCharaRect(int code)
        {
            Rectangle rc = new Rectangle()
            {
                X = code % 16 * charaSize + charaBaseX,
                Y = code / 16 * charaSize + charaBaseY,
                Width = charaSize,
                Height = charaSize,
            };
            return rc;
        }

        // -------------------------------------------------------------
        // マウスカーソルを表示する
        // -------------------------------------------------------------
        private void DrawMouseCursor(Graphics e)
        {
            // マウスカーソル表示
            int code = GetCharaCode(PointToClient(Cursor.Position));
            Rectangle rc = GetCharaRect(code);
            rc.Width++;
            rc.Height++;
            e.DrawRectangle(penCursor, rc);
        }

        // -------------------------------------------------------------
        // 選択カーソルを表示する
        // -------------------------------------------------------------
        private void DrawSelectedCursor(Graphics e)
        {
            // 編集カーソル表示
            if (numColor < 4) return;
            Rectangle rc = GetCharaRect(chr.code);
            rc.Width++;
            rc.Height++;
            e.FillRectangle(Brushes.Gray, rc);
        }

        // -------------------------------------------------------------
        // 再描画される
        // -------------------------------------------------------------
        private void DlgParts_Paint(object sender, PaintEventArgs e)
        {
            // 透明部分を消す
            Color backColor = (int)chr.color > 2 ? Color.Black : Color.White;
            e.Graphics.Clear(backColor);
            Rectangle rcErase = new Rectangle(charaBaseX, charaBaseY, 12, 12);
            MainForm.FillTransparent(e.Graphics, rcErase);

            // キャラ文字を表示する
            Rectangle rcDest = new Rectangle(charaBaseX, charaBaseY, 256, 256);
            ImageAttributes ia = MainForm.GetImageAtribute(chr.color);
            e.Graphics.DrawImage(pc80chara, rcDest, 0, 0, 256, 256, GraphicsUnit.Pixel, ia);

            // カーソルを表示する
            DrawSelectedCursor(e.Graphics);
            DrawMouseCursor(e.Graphics);

            // 説明文の色を変更する
            labelHelp.BackColor = backColor;
            labelHelp.ForeColor = backColor == Color.White ? Color.Black : Color.White;
        }

        // -------------------------------------------------------------
        // マウスが移動している
        // -------------------------------------------------------------
        private void DlgParts_MouseMove(object sender, MouseEventArgs e)
        {
            Invalidate();
        }

        // -------------------------------------------------------------
        // マウスクリックされた
        // -------------------------------------------------------------
        private void DlgParts_MouseClick(object sender, MouseEventArgs e)
        {
            switch (e.Button)
            {
                case MouseButtons.Left:
                    chr.code = GetCharaCode(PointToClient(Cursor.Position));
                    EndDlg();
                    break;
                case MouseButtons.Right:
                    break;
            }
        }

        // -------------------------------------------------------------
        // テキストボックスの入力内容が変化した
        // -------------------------------------------------------------
        private void textBox_TextChanged(object sender, EventArgs e)
        {
            if (reEdnterTextBox) return;
            reEdnterTextBox = true;

            // キャラ選択を変更する
            if (textBox.Text.Length > 0)
            {
                int code = textBox.Text[0];
                if (chr.code == textBox.Text[0])
                {
                    EndDlg();
                    return;
                }
                chr.code = textBox.Text[0];
                textBox.Text = "";
                Invalidate();
            }

            reEdnterTextBox = false;
        }

        // -------------------------------------------------------------
        // 変更を確定してダイアログを閉じる
        // -------------------------------------------------------------
        private void EndDlg()
        {
            if (front) MainForm.form.SetSelectChara(chr);
            else MainForm.form.SetBackChara(chr);
            Close();
        }

        // -------------------------------------------------------------
        // フォームでキーが押された
        // -------------------------------------------------------------
        private void DlgParts_KeyUp(object sender, KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.Up:
                    if (chr.code >= 16) chr.code -= 16;
                    Invalidate();
                    break;
                case Keys.Down:
                    if (chr.code < 256 - 16) chr.code += 16;
                    Invalidate();
                    break;
                case Keys.Left:
                    if (chr.code > 0) --chr.code;
                    Invalidate();
                    break;
                case Keys.Right:
                    if (chr.code < 255) ++chr.code;
                    Invalidate();
                    break;
                case Keys.Enter:
                    EndDlg();
                    break;
                case Keys.Escape:
                    Close();
                    break;

            }
        }
    }
}
