﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Runtime.Serialization.Formatters.Binary;
using System.Windows.Forms;
using Microsoft.WindowsAPICodePack;
using Microsoft.WindowsAPICodePack.Dialogs;
using Microsoft.WindowsAPICodePack.Dialogs.Controls;
using static CharaEdit40.Define;

namespace CharaEdit40
{
    public partial class MainForm : Form
    {
        public static MainForm form;

        public List<Layer> lsCharas = new List<Layer>();                // キャラテーブル（レイヤ付き）
        private int editLayer = 0;                                      // 編集対象レイヤー
        private Point posCursor = new Point();                          // 編集場所

        private List<UndoData> lsUndo = new List<UndoData>();           // アンドゥバッファ
        private int undoPos = 0;                                        // アンドゥポインタ（リドゥで使用）
        private bool enableUndo = true;                                 // アンドゥ追加可能か

        private string productTitle = String.Empty;                     // 製品名
        private String filePath = String.Empty;                         // ファイル名
        private Bitmap pc80Font = Properties.Resources._80font;         // フォント画像
        private Bitmap checkedEye = Properties.Resources.checkedEye;    // 視認アイコン画像
        private Bitmap bmpScreen = null;                                // 編集画像
        private Bitmap layerThumbBack = null;                           // レイヤーサムネイルの背景画像
        private Bitmap[] atrbMarkers = null;                            // アトリビュート警告マーカー
        private List<int> lsAtrb = new List<int>();                     // アトリビュートラインデータリスト
        private int atrbPrnLine = -1;                                   // アトリビュートカウント表示ライン

        private bool reEnterListView = false;                           // リストビュー再入禁止フラグ
        private bool reEdnterTextBox = false;                           // テキストボックス再入禁止フラグ
        private bool editLabel = false;                                 // ラベル編集中か
        private bool enableDrawScreen = true;                           // 編集エリアが再描画可能か
        private bool alreadyEdited = false;                             // 編集済みか
        private bool wrnAtrbNotBeginLine = false;                       // 特殊Atrbは行頭設定不可警告を表示済みか
        private bool wrnAtrbNotNullCode = false;                        // 特殊Atrbは透明には設定不能警告を表示済みか

        private DlgParts dlgParts = null;
        private static Brush brGray = new SolidBrush(Color.FromArgb(204, 204, 204));
        private static Color clGrid = Color.FromArgb(130, 130, 130);
        private static Pen penGrid = new Pen(clGrid);
        private static Pen penWhite = new Pen(Color.White);
        private static Pen penGridLine = new Pen(Color.Black, 2);
        private static Pen penGridSelect = new Pen(Color.White, 2);
        private static Brush brLVItem = new SolidBrush(Color.FromArgb(94, 94, 94));
        private static Brush brLVItemSelect = new SolidBrush(Color.FromArgb(30, 30, 30));
        public static Color[] penColors = new Color[]
        {
            Color.Black,
            Color.Blue, Color.Red, Color.Magenta, Color.Lime,
            Color.Cyan, Color.Yellow, Color.White
        };
        public static Brush[] brColors = new Brush[]
        {
            new SolidBrush(Color.FromArgb(128, Color.Black)),
            new SolidBrush(Color.FromArgb(128, Color.Blue)),
            new SolidBrush(Color.FromArgb(128, Color.Red)),
            new SolidBrush(Color.FromArgb(128, Color.Magenta)),
            new SolidBrush(Color.FromArgb(128, Color.Lime)),
            new SolidBrush(Color.FromArgb(128, Color.Cyan)),
            new SolidBrush(Color.FromArgb(128, Color.Yellow)),
            new SolidBrush(Color.FromArgb(128, Color.White)),
        };
        private static Color[] alartColors = new Color[]
        {
                Color.FromArgb(  0, 255, 0),
                Color.FromArgb(  0, 255, 0),
                Color.FromArgb(  0, 255, 0),
                Color.FromArgb(  0, 255, 0),
                Color.FromArgb(  0, 255, 0),
                Color.FromArgb(  0, 255, 0),

                Color.FromArgb( 16, 255, 0),
                Color.FromArgb( 32, 255, 0),
                Color.FromArgb( 64, 255, 0),
                Color.FromArgb(128, 255, 0),
                Color.FromArgb(255, 255, 0),

                Color.FromArgb(255, 192, 0),
                Color.FromArgb(255, 128, 0),
                Color.FromArgb(255,  96, 0),
                Color.FromArgb(255,  64, 0),
                Color.FromArgb(255,  32, 0),
        };
        private static byte[] pc8001atrbs = new byte[]
        {
            0x08, 0x28, 0x48, 0x68, 0x88, 0xA8, 0xC8, 0xE8
        };
        private static byte[] atrbs = new byte[]
        {
            ATB_TRANSPARENT, ATB_BLACK, ATB_BLUE, ATB_RED, ATB_MAGENTA, ATB_GREEN, ATB_CYAN, ATB_YELLOW, ATB_WHITE
        };
        private static float[] zoomRates = new float[]
        {
            0.1f, 0.2f, 0.5f, 1.0f, 2.0f, 3.0f, 4.0f, 5.0f, 6.0f, 7.0f, 8.0f, 9.0f, 10.0f
        };

        // 選択キャラ文字
        public static Chara charaSelect = new Chara();                  // 選択キャラ
        public static Chara charaBack = new Chara();                    // 待機キャラ
        private Timer timerCursor = new Timer();                        // 選択枠カーソル点滅タイマー
        private uint timerCounter = 0;                                  // 汎用カウンター
        private bool drawCursor = true;                                 // カーソル表示フラグ

        // 選択範囲
        private Rectangle rcSelect = new Rectangle();                   // 選択範囲
        private Point offsetSelect = new Point();                       // 選択開始座標
        private Pen[] penSelects = new Pen[]                            // 選択枠
        {
            new Pen(Color.White, 2),
            new Pen(Color.Black, 2),
        };
        private Timer timerSelect = new Timer();                        // 選択枠アニメタイマー

        // 移動
        private Block moveBlock = new Block();                          // 移動するデータブロック

        // コピーペースト
        private Block pasteBlock = new Block();                         // ペースト用データバッファ

        // タトリビュート警告
        private PictureBox picAtrbMarker = new PictureBox();            // アトリビュートマーカーコントロール          

        // コマンド選択
        private Label lbCommand = new Label();                          // コマンド表示テキスト
        private Cmd mode = Cmd.Pen;                                     // 現在の動作モード

        // ズーム
        private int zoomCnt = 4;                                        // ズームレートカウンター
        private TextBox zoomText = new TextBox();                       // 表示用テキストコントロール

        // レイヤーリスト
        private BufferedListView lvLayer = new BufferedListView();

        // 情報表示
        private Label lbInformation = new Label();                      // 各種情報表示

        // ステータスバー
        private PictureBox picCommand = new PictureBox();               // 選択コマンドアイコン画像コントロール
        private PictureBox picSelect = new PictureBox();                // 選択キャラ画像コントロール
        private PictureBox picBack = new PictureBox();                  // 待機キャラ画像コントロール
        private ToolStripControlHost[] stlpTools;                       // ステータスバーアイテム配列
        
        // -------------------------------------------------------------
        // コンストラクタ
        // -------------------------------------------------------------
        public MainForm()
        {
            InitializeComponent();

            AssemblyProductAttribute asmprd = (AssemblyProductAttribute)Attribute.GetCustomAttribute
            (
                Assembly.GetExecutingAssembly(), typeof(AssemblyProductAttribute)
            );
            productTitle = asmprd.Product;

            atrbMarkers = new Bitmap[]
            {
                Properties.Resources.markGreen,
                Properties.Resources.markYellow,
                Properties.Resources.markRed,
                Properties.Resources.markOut,
            };

            form = this;
            enableDrawScreen = true;
            editLabel = false;
        }

        // -------------------------------------------------------------
        // lvLayerを初期化する
        // -------------------------------------------------------------
        private void InitLvLayer()
        {
            lvLayer.Alignment = ListViewAlignment.Left;
            lvLayer.Anchor = AnchorStyles.Top | AnchorStyles.Bottom | AnchorStyles.Left | AnchorStyles.Right;
            lvLayer.BackColor = Color.FromArgb(94, 94, 94);
            lvLayer.CheckBoxes = true;
            lvLayer.ContextMenuStrip = ctxLayerMenu;
            lvLayer.ForeColor = SystemColors.MenuText;
            lvLayer.FullRowSelect = true;
            lvLayer.GridLines = true;
            lvLayer.HeaderStyle = ColumnHeaderStyle.None;
            lvLayer.HideSelection = false;
            lvLayer.LabelEdit = true;
            lvLayer.LargeImageList = imgListLayer;
            lvLayer.Location = new Point(5, 3);
            lvLayer.MultiSelect = false;
            lvLayer.Name = "lvLayer";
            lvLayer.OwnerDraw = true;
            lvLayer.Size = new Size(206, 414);
            lvLayer.SmallImageList = imgListLayer;
            lvLayer.TabIndex = 1;
            lvLayer.UseCompatibleStateImageBehavior = false;
            lvLayer.View = View.List;
            lvLayer.AfterLabelEdit += lvLayer_AfterLabelEdit;
            lvLayer.BeforeLabelEdit += lvLayer_BeforeLabelEdit;
            lvLayer.DrawItem += lvLayer_DrawItem;
            lvLayer.ItemChecked += lvLayer_ItemChecked;
            lvLayer.SelectedIndexChanged += lvLayer_SelectedIndexChanged;
            splitMain.Panel2.Controls.Add(lvLayer);
        }

        // -------------------------------------------------------------
        // フォームがロードされる
        // -------------------------------------------------------------
        private void MainForm_Load(object sender, EventArgs e)
        {

            enableUndo = false;

            //分割線を固定する
            splitMain.IsSplitterFixed = true;

            // レイヤーリストを初期化する
            InitLvLayer();

            // ダブルバッファON
            DoubleBuffered = true;
            SetStyle(ControlStyles.ResizeRedraw, true);

            // カーソル点滅
            timerCursor.Tick += OnTickCursor;
            timerCursor.Interval = 100;
            timerCursor.Start();

            // 選択ライン
            penSelects[0].DashPattern = penSelects[1].DashPattern = new float[] { 4f, 4f };
            penSelects[0].DashOffset = 0;
            penSelects[1].DashOffset = 4;
            timerSelect.Tick += OnTickSelectAnime;
            timerSelect.Interval = 100;
            timerSelect.Start();

            // ステータスバー表示用画像を初期化する
            picCommand.Name = "picCommand";
            picCommand.Image = Properties.Resources.CmdPen;
            picCommand.MouseClick += LabelCommand_Click;

            lbCommand.TextAlign = ContentAlignment.MiddleLeft;
            lbCommand.AutoSize = false;
            lbCommand.Size = new Size(60, 24);
            lbCommand.MouseClick += LabelCommand_Click;

            picSelect.Name = "picSelected";
            picSelect.Bounds = new Rectangle(0, 0, SELICON_WIDTH, SELICON_HEIGHT);
            picSelect.Image = new Bitmap(SELICON_WIDTH, SELICON_HEIGHT);
            picSelect.MouseClick += picSelect_MouseClick;
            picBack.Name = "picBack";
            picBack.Bounds = new Rectangle(0, 0, SELICON_WIDTH, SELICON_HEIGHT);
            picBack.Image = new Bitmap(SELICON_WIDTH, SELICON_HEIGHT);
            picBack.MouseClick += picBack_MouseClick;

            // ステータスバー表示用ズームレートテキストを初期化する
            zoomText.TextAlign = HorizontalAlignment.Left;
            zoomText.Anchor = AnchorStyles.Left;
            zoomText.Text = string.Format("{0, -8}", zoomRates[zoomCnt].ToString("P1"));
            zoomText.ReadOnly = true;

            // ステータスバー表示用表示プレートを初期化する
            lbInformation.TextAlign = ContentAlignment.MiddleLeft;
            lbInformation.BorderStyle = BorderStyle.None;
            lbInformation.Visible = false;

            // ステータスバーを初期化する
            stlpTools = new ToolStripControlHost[]
            {
                new ToolStripControlHost(picCommand),
                new ToolStripControlHost(lbCommand),
                new ToolStripControlHost(picSelect),
                new ToolStripControlHost(picBack),
                new ToolStripControlHost(zoomText),
                new ToolStripControlHost(lbInformation),
            };
            stlpTools[0].AutoSize = false;
            int statusPos = 0;
            statusStripMain.Items.Add(stlpTools[statusPos++]);
            statusStripMain.Items.Add(stlpTools[statusPos++]);
            statusStripMain.Items.Add(new ToolStripSeparator());
            statusStripMain.Items.Add(stlpTools[statusPos++]);
            statusStripMain.Items.Add(stlpTools[statusPos++]);
            statusStripMain.Items.Add(new ToolStripSeparator());
            statusStripMain.Items.Add(stlpTools[statusPos++]);
            statusStripMain.Items.Add(new ToolStripSeparator());
            statusStripMain.Items.Add(stlpTools[statusPos++]);

            statusStripMain.Items[0].AutoSize = false;
            statusStripMain.Items[0].Size = new Size(24, 24);
            statusStripMain.Items[0].ImageAlign = ContentAlignment.MiddleRight;
            statusStripMain.Items[0].Margin = new Padding(10, 3, 0, 0);
            statusStripMain.Items[1].AutoSize = false;
            PrintCommand();

            // 画面を初期化する
            charaSelect.code = 0x87;    // ■
            charaSelect.color = ExColor.Green;
            DrawStatusCharactor(picSelect, charaSelect);
            charaBack.code = 0;
            charaBack.color = ExColor.Red;
            DrawStatusCharactor(picBack, charaBack);

            // 編集エリアにアトリビュート警告マーカーを追加する
            splitMain.Panel1.Controls.Add(picAtrbMarker);
            picAtrbMarker.BackColor = splitMain.Panel1.BackColor;
            picAtrbMarker.Visible = false;

            // 画像データを初期化する
            bmpScreen = new Bitmap(TEXT_WIDTH * CHIP_SIZE, TEXT_HEIGHT * CHIP_SIZE);
            CreatelayerThumbBack();

            // コマンドラインをプロジェクトファイルとみなして読み込む
            string[] cmds = System.Environment.GetCommandLineArgs();
            for (int i = 1; i < cmds.Count(); ++i)
            {
                if (i == 1) FileOpen(cmds[i]);
                else        FileMerge(cmds[i]);
            }
            if (filePath == String.Empty)
            {
                // 新規レイヤ状態とする
                FileNewToolStripMenuItem_Click(null, null);

            }

            // タイトルを更新する
            enableUndo = true;
            Icon = Properties.Resources.icon256;
            PrintTitle();
            SetCursorCenter();
        }

        // -------------------------------------------------------------
        // フォームが表示される
        // -------------------------------------------------------------
        private void MainForm_Shown(object sender, EventArgs e)
        {
            picScreen.Focus();
        }
        // -------------------------------------------------------------
        // カーソル位置をセンターに設定する
        // -------------------------------------------------------------
        private void SetCursorCenter()
        {
            Size sz = GetChrSize();
            posCursor = new Point(sz.Width / 2, sz.Height / 2);
        }

        // -------------------------------------------------------------
        // カーソル点滅タイマー
        // -------------------------------------------------------------
        private void OnTickCursor(object sender, EventArgs e)
        {
            ++timerCounter;
            if (timerCounter % 5 == 0) drawCursor = !drawCursor;
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // 選択枠アニメーションタイマー
        // -------------------------------------------------------------
        private void OnTickSelectAnime(object sender, EventArgs e)
        {
            if (rcSelect.Width == 0) return;
            ++penSelects[0].DashOffset;
            ++penSelects[1].DashOffset;
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // 何かキーが押された（フック）
        // -------------------------------------------------------------
        [System.Security.Permissions.UIPermission(
            System.Security.Permissions.SecurityAction.Demand,
            Window = System.Security.Permissions.UIPermissionWindow.AllWindows)]
        protected override bool ProcessDialogKey(Keys keyData)
        {
            if (editLabel || mode == Cmd.Text) return false;

            bool keyCtrl = (ModifierKeys & Keys.Control) == Keys.Control;
            bool keyShift = (ModifierKeys & Keys.Shift) == Keys.Shift;
            bool keyAlt = (ModifierKeys & Keys.Alt) == Keys.Alt;

            // 直接コマンド
            switch (keyData & Keys.KeyCode)
            {
                case Keys.Z:        // [Z] キャラ配置
                    OnClickPen();
                    return true;
                case Keys.X:        // [X] キャラ文字入れ替え
                    Chara tmp = charaSelect;
                    charaSelect = charaBack;
                    charaBack = tmp;
                    DrawStatusCharactor(picSelect, charaSelect);
                    DrawStatusCharactor(picBack, charaBack);
                    picSelect.Invalidate();
                    picBack.Invalidate();
                    return true;
                case Keys.C:        // [C] 選択キャラ文字選択
                    OnChangeChara(true);
                    return true;
                case Keys.B:        // [B] バックキャラ文字選択
                    OnChangeChara(false);
                    return true;
                case Keys.M:        // [M] 長方形選択
                    SetCommand(Cmd.Select);
                    return true;
                case Keys.P:        // [P] 鉛筆
                    EditCharaSetToolStripMenuItem_Click(null, null);
                    SetCommand(Cmd.Pen);
                    return true;
                case Keys.V:        // [V] 移動
                    SetCommand(Cmd.Move);
                    return true;
                case Keys.I:        // [I] テキスト入力
                    OnClickText();
                    break;
                case Keys.Delete:   // [DEL] 範囲削除
                    if (!keyCtrl)
                    {
                        if (picScreen.Focused && rcSelect.Width > 0)
                        {
                            EditFillRectToolStripMenuItem_Click(null, null);
                            return true;
                        }
                    }
                    break;
                case Keys.R:        // [R] Atrb:反転指定
                    SetAtrbReverse();
                    return true;
                case Keys.U:        // [U] Atrb:上線指定
                    SetAtrbOverLine();
                    return true;
                case Keys.D:        // [D] Atrb:下線指定
                    SetAtrbUnderLine();
                    return true;
                case Keys.N:        // [N] Atrb:特殊設定削除
                    SetAtrbRemove();
                    return true;
                case Keys.PageDown:
                    if (!ViewAtrbToolStripMenuItem.Checked || lsCharas.Count <= 0) break;
                    if (atrbPrnLine >= lsCharas[0].chrTable.GetLength(ELEMENT_HEIGHT) - 1) break;
                    if (keyCtrl)
                    {
                        // [Ctrl]+[PGUP] 今より変化点の少ないラインに移動する
                        DrawAtrbImage(null);
                        int cnt = lsAtrb[atrbPrnLine];
                        int bak = atrbPrnLine;
                        for (--cnt; cnt >= 0; --cnt)
                        {
                            atrbPrnLine = lsAtrb.IndexOf(cnt);
                            if (atrbPrnLine >= 0) break;
                        }
                        if (cnt < 0) atrbPrnLine = bak;
                    }
                    else
                    {
                        // [PGDN] 表示アトリビュートを1ライン下げる / +[Shift] 10ライン下げる
                        atrbPrnLine += keyShift ? 10 : 1;
                        if (atrbPrnLine > lsCharas[0].chrTable.GetLength(ELEMENT_HEIGHT) - 1)
                        {
                            atrbPrnLine = lsCharas[0].chrTable.GetLength(ELEMENT_HEIGHT) - 1;
                        }
                    }
                    picScreen.Invalidate();
                    return true;
                case Keys.PageUp:
                    if (!ViewAtrbToolStripMenuItem.Checked || lsCharas.Count <= 0 || atrbPrnLine < 0) break;
                    if (keyCtrl)    
                    {
                        // [Ctrl]+[PGUP] 今より変化点の多いラインに移動する
                        DrawAtrbImage(null);
                        int cnt = lsAtrb[atrbPrnLine];
                        for (++cnt; cnt <= lsAtrb.Max(); ++cnt)
                        {
                            atrbPrnLine = lsAtrb.IndexOf(cnt);
                            if (atrbPrnLine >= 0) break;
                        }
                    }
                    else
                    {
                        // [PGDN] 表示アトリビュートを1ライン上げる / +[Shift] 10ライン上げる
                        atrbPrnLine -= keyShift ? 10 : 1;
                        if (atrbPrnLine < 0) atrbPrnLine = 0;
                    }
                    picScreen.Invalidate();
                    return true;
                case Keys.Home:     // [Home] 表示アトリビュートを一番上に / +[Ctrl] 一番変化の多い場所に
                    if (!ViewAtrbToolStripMenuItem.Checked || lsCharas.Count <= 0) break;
                    atrbPrnLine = keyCtrl ? -1 : 0;
                    picScreen.Invalidate();
                    return true;
                case Keys.End:      // [End] 表示アトリビュートを一番下に / +[Ctrl] 一番変化の少ない場所に
                    if (!ViewAtrbToolStripMenuItem.Checked || lsCharas.Count <= 0) break;
                    if (keyCtrl)
                    {
                        DrawAtrbImage(null);
                        for (int cnt = 0; cnt <= lsAtrb.Max(); ++cnt)
                        {
                            atrbPrnLine = lsAtrb.IndexOf(cnt);
                            if (atrbPrnLine >= 0) break;
                        }
                    }
                    else
                    {
                        atrbPrnLine = lsCharas[0].chrTable.GetLength(ELEMENT_HEIGHT) - 1;
                    }
                    picScreen.Invalidate();
                    return true;
                case Keys.Oemplus:
                case Keys.Add:      // [Ctrl]+[+]
                    if (!keyCtrl) break;
                    ViewZoomInToolStripMenuItem_Click(null, null);
                    return true;
                case Keys.OemMinus:
                case Keys.Subtract: // [Ctrl]+[-]
                    if (!keyCtrl) break;
                    ViewZoomOutToolStripMenuItem_Click(null, null);
                    return true;
                case Keys.D0:       // [0] 黒 / [Shift]+[0] 透明 / [Ctrl]+[0] 画面サイズに合わせる
                case Keys.NumPad0:
                    if (keyCtrl)
                    {
                        if (keyAlt) ImageCanvasFitToolStripMenuItem_Click(null, null);   // 編集サイズに合わせる
                        else ViewZoomFitToolStripMenuItem_Click(null, null);            // 画面サイズに合わせる
                    }
                    else
                    {
                        if (keyShift) SetSelectChara(0, charaSelect.color);             // 透明パターン
                        else if (keyAlt) RepTableColor(ExColor.Black);                  // 黒色変更
                        else SetSelectChara(charaSelect.code, ExColor.Black);           // 黒
                    }
                    return true;
                case Keys.D1:       // [1] 青 / [Alt]+[1] 青色に置換 / [ctrl]+[1] ピクセル等価
                case Keys.NumPad1:
                    if (keyCtrl) PixelEquToolStripMenuItem_Click(null, null);
                    else if (keyAlt) RepTableColor(ExColor.Blue);
                    else SetSelectChara(charaSelect.code, ExColor.Blue);
                    return true;
                case Keys.D2:       // [2] 赤
                case Keys.NumPad2:
                    if (keyCtrl) break;
                    if (keyAlt) RepTableColor(ExColor.Red);
                    else SetSelectChara(charaSelect.code, ExColor.Red);
                    return true;
                case Keys.D3:       // [3] 紫
                case Keys.NumPad3:
                    if (keyCtrl) break;
                    if (keyAlt) RepTableColor(ExColor.Magenta);
                    else SetSelectChara(charaSelect.code, ExColor.Magenta);
                    return true;
                case Keys.D4:       // [4] 緑
                case Keys.NumPad4:
                    if (keyCtrl) break;
                    if (keyAlt) RepTableColor(ExColor.Green);
                    else SetSelectChara(charaSelect.code, ExColor.Green);
                    return true;
                case Keys.D5:       // [5] 水色
                case Keys.NumPad5:
                    if (keyCtrl) break;
                    if (keyAlt) RepTableColor(ExColor.Cyan);
                    else SetSelectChara(charaSelect.code, ExColor.Cyan);
                    return true;
                case Keys.D6:       // [6] 黄
                case Keys.NumPad6:
                    if (keyCtrl) break;
                    if (keyAlt) RepTableColor(ExColor.Yellow);
                    else SetSelectChara(charaSelect.code, ExColor.Yellow);
                    return true;
                case Keys.D7:       // [7] 白
                case Keys.NumPad7:
                    if (keyCtrl) break;
                    if (keyAlt) RepTableColor(ExColor.White);
                    else SetSelectChara(charaSelect.code, ExColor.White);
                    return true;
            }
            return false;
        }

        // -------------------------------------------------------------
        // 特殊アトリビュート指定を選択枠／カーソル位置に付与する
        // -------------------------------------------------------------
        private void SetAtribute(int atrb)
        {
            Rectangle rc = new Rectangle(new Point(), GetChrSize());
            Rectangle rcSrc = rcSelect.IsEmpty
                ? new Rectangle(GetMapPos(picScreen.PointToClient(Cursor.Position)), new Size(1, 1))
                : rcSelect;
            rc.Intersect(rcSrc);
            if (rc.IsEmpty) return;

            // そのレイヤーは見えているか
            switch (CheckShowLayer())
            {
                case DialogResult.Yes:
                    lvLayer.Items[editLayer].Checked = true;
                    return;     // ダイアログ選択で位置が変わるのでキャラ配置は一旦キャンセル
                case DialogResult.No:
                    return;
            }

            SetUndo();
            for (int y = 0; y < rc.Height; ++y)
            {
                for (int x = 0; x < rc.Width; ++x)
                {
                    int posX = rc.X + x;
                    int posY = rc.Y + y;
                    if (posX <= 0)
                    {
                        if (!wrnAtrbNotBeginLine)
                        {
                            MessageBox.Show(
                                text: "行頭に特殊アトリビュートは設定できません。",
                                caption: ProductName,
                                buttons: MessageBoxButtons.OK,
                                icon: MessageBoxIcon.Information
                            );
                            wrnAtrbNotBeginLine = true;
                        }
                        lsCharas[editLayer].chrTable[0, posY].atrb = ATB_NORMAL;
                        continue;
                    }
                    if (lsCharas[editLayer].chrTable[posX, posY].code == 0)
                    {
                        if (!wrnAtrbNotNullCode)
                        {
                            bool anotherLayer = false;
                            for (int i = 0; !anotherLayer && i < lsCharas.Count; ++i)
                            {
                                if (i == editLayer) continue;
                                anotherLayer = lsCharas[i].chrTable[posX, posY].code != 0;
                            }
                            MessageBox.Show(
                                text: "透明部分に特殊アトリビュートは設定できません。"
                                    + (anotherLayer ? Environment.NewLine + "もしかして選択レイヤーを間違えていませんか？" : ""),
                                caption: ProductName,
                                buttons: MessageBoxButtons.OK,
                                icon: MessageBoxIcon.Information
                            );
                            wrnAtrbNotNullCode = true;
                        }
                        continue;
                    }
                    if (atrb == ATB_NORMAL)
                    {
                        // 特殊アトリビュートを消す
                        lsCharas[editLayer].chrTable[posX, posY].atrb = atrb;
                    }
                    else
                    {
                        // 特殊アトリビュートをビット反転する
                        lsCharas[editLayer].chrTable[posX, posY].atrb ^= atrb;
                    }
                }
            }
            CreateLayerPicture(lsCharas[editLayer]);
            lvLayer.Refresh();
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // アトリビュート反転指定を選択枠／カーソル位置に付与する
        // -------------------------------------------------------------
        private void SetAtrbReverse()
        {
            SetAtribute(ATB_REVERSE);
        }

        // -------------------------------------------------------------
        // アトリビュート上線指定を選択枠／カーソル位置に付与する
        // -------------------------------------------------------------
        private void SetAtrbOverLine()
        {
            SetAtribute(ATB_OVERLINE);
        }

        // -------------------------------------------------------------
        // アトリビュート下線指定を選択枠／カーソル位置に付与する
        // -------------------------------------------------------------
        private void SetAtrbUnderLine()
        {
            SetAtribute(ATB_UNDERLINE);
        }

        // -------------------------------------------------------------
        // 特殊アトリビュートを選択枠／カーソル位置から削除する
        // -------------------------------------------------------------
        private void SetAtrbRemove()
        {
            SetAtribute(ATB_NORMAL);
        }

        // -------------------------------------------------------------
        // カーソル形状を設定する
        // -------------------------------------------------------------
        private void SetCursorIcon()
        {
            switch (mode)
            {
                case Cmd.Pen:
                case Cmd.Move:
                    picScreen.Cursor = IsCurosrSelectRange() ? Cursors.Default : Cursors.Cross;
                    break;
                case Cmd.Select:
                case Cmd.SelectMove:
                    if (GetRectSelect().IsEmpty) picScreen.Cursor = Cursors.Default;
                    else picScreen.Cursor = IsCurosrSelectRange() ? Cursors.SizeAll : Cursors.Cross;
                    break;
                case Cmd.Text:
                    picScreen.Cursor = Cursors.IBeam;
                    break;
                default:
                    picScreen.Cursor = Cursors.Default;
                    break;
            }
        }

        // -------------------------------------------------------------
        // 現在の動作モードを変更する
        // -------------------------------------------------------------
        private void SetCommand(Cmd c)
        {
            switch (mode = c)
            {
                case Cmd.Pen:
                case Cmd.PenOn:
                    EditPenToolStripMenuItem.Checked = true;
                    EditSelectToolStripMenuItem.Checked = false;
                    EditMoveToolStripMenuItem.Checked = false;
                    EditTextInputToolStripMenuItem.Checked = false;
                    break;
                case Cmd.Select:
                    EditPenToolStripMenuItem.Checked = false;
                    EditSelectToolStripMenuItem.Checked = true;
                    EditMoveToolStripMenuItem.Checked = false;
                    EditTextInputToolStripMenuItem.Checked = false;
                    break;
                case Cmd.Move:
                    EditPenToolStripMenuItem.Checked = false;
                    EditSelectToolStripMenuItem.Checked = false;
                    EditMoveToolStripMenuItem.Checked = true;
                    EditTextInputToolStripMenuItem.Checked = false;
                    break;
                case Cmd.Text:
                    EditPenToolStripMenuItem.Checked = false;
                    EditSelectToolStripMenuItem.Checked = false;
                    EditMoveToolStripMenuItem.Checked = false;
                    EditTextInputToolStripMenuItem.Checked = true;
                    break;
            }
            SetCursorIcon();
            PrintCommand();
        }

        // -------------------------------------------------------------
        // アプリタイトルを更新する
        // -------------------------------------------------------------
        private void PrintTitle()
        {
            string filename = Path.GetFileNameWithoutExtension(filePath);
            if (filename == String.Empty) filename = "(無題)";
            else filename += " ";
            if (alreadyEdited) filename += "(更新)";
            filename += " - " + productTitle;
            MainForm.form.Text = filename;
        }

        // -------------------------------------------------------------
        // アンドゥを追加する
        // -------------------------------------------------------------
        private void SetUndo()
        {
            if (!enableUndo) return;
            if (lsUndo.Count != undoPos )
            {
                lsUndo.RemoveRange(undoPos, lsUndo.Count - undoPos);
            }
            UndoData buf = new UndoData()
            {
                lsCharas = new List<Layer>(),
                editLayer = editLayer,
                filename = filePath,
                posCursor = posCursor
            };
            for (int i = 0; i < lsCharas.Count; ++i)
            {
                buf.lsCharas.Add(new Layer(lsCharas[i]));
            }
            lsUndo.Add(buf);
            undoPos = lsUndo.Count;
            alreadyEdited = true;
            PrintTitle();
        }

        // -------------------------------------------------------------
        // 選択キャラ文字を変更する
        // -------------------------------------------------------------
        public void SetSelectChara(Chara chr)
        {
            SetSelectChara(chr.code, chr.color,chr.atrb);
        }
        public void SetSelectChara(int code, ExColor color, int atrb = 0)
        {
            charaSelect.code = code;
            charaSelect.color = color;
            charaSelect.atrb = atrb;
            DrawStatusCharactor(picSelect, charaSelect);
            picSelect.Invalidate();
        }

        // -------------------------------------------------------------
        // バックキャラ文字を変更する
        // -------------------------------------------------------------
        public void SetBackChara(Chara chr) { SetbackChara(chr.code, chr.color, chr.atrb); }
        public void SetbackChara(int code, ExColor color, int atrb = 0)
        {
            charaBack.code = code;
            charaBack.color = color;
            charaBack.atrb = atrb;
            DrawStatusCharactor(picBack, charaBack);
            picBack.Invalidate();
        }

        // -------------------------------------------------------------
        // 現在の動作モードを表示する
        // -------------------------------------------------------------
        private void PrintCommand()
        {
            string str = string.Empty;
            string desc = string.Empty;
            switch (mode)
            {
                case Cmd.Pen:
                case Cmd.PenOn:
                    picCommand.Image = Properties.Resources.CmdPen;
                    str = "鉛筆";
                    desc = " [↑,↓,←,→]カーソル移動　[Z]選択文字配置　[Shift]+[Z]バック文字配置　[C]選択文字変更　[B]バック文字変更";
                    break;
                case Cmd.Move:
                    picCommand.Image = Properties.Resources.CmdMove;
                    str = "移動";
                    desc = " マウスドラッグで範囲を移動します。キーボード使用不可。";
                    break;
                case Cmd.MoveOn:
                    picCommand.Image = Properties.Resources.CmdMove;
                    str = "移動";
                    desc = " マウスクリックを離すと移動先で確定します。";
                    break;
                case Cmd.Select:
                    picCommand.Image = Properties.Resources.CmdSelect;
                    str = "範囲選択";
                    desc = " マウスで範囲を指定します。キーボード使用不可。";
                    break;
                case Cmd.SelectOn:
                    picCommand.Image = Properties.Resources.CmdSelect;
                    str = "範囲選択";
                    desc = " マウスクリックを離すと範囲が確定します。";
                    break;
                case Cmd.SelectMove:
                    picCommand.Image = Properties.Resources.CmdSelect;
                    str = "範囲選択";
                    desc = " マウスクリックを離すと範囲の位置変更が確定します。";
                    break;
                case Cmd.DlgPattern:
                    picCommand.Image = Properties.Resources.DlgParts;
                    str = "文字選択";
                    desc = " [Enter]選択文字確定 [Esc]選択文字変更キャンセル。";
                    break;
                case Cmd.Text:
                    picCommand.Image = Properties.Resources.CmdText;
                    str = "文字入力";
                    desc = " [↑,↓,←,→]カーソル移動　[Z]選択文字配置　[Shift]+[Z]バック文字配置　[C]選択文字変更　[B]バック文字変更";
                    break;
            }
            lbCommand.Text = str;
            if (!ViewAtrbToolStripMenuItem.Checked)
            {
                lbInformation.Text = desc;
                lbInformation.BorderStyle = BorderStyle.None;
                lbInformation.BackColor = SystemColors.Control;
                lbInformation.ForeColor = SystemColors.ControlText;
            }
            lbInformation.Visible = true;
        }

        // -------------------------------------------------------------
        // ExColor 色コードから Colorに変換する
        // -------------------------------------------------------------
        public static Color GetColor(ExColor color)
        {
            return color == ExColor.Transparent ? Color.Transparent : penColors[(int)color];
        }

        // -------------------------------------------------------------
        // Colorコードから ExColorに変換する
        // -------------------------------------------------------------
        public static ExColor GetColor(Color color)
        {
            if (color.A < 128) return ExColor.Transparent;
            int col = (color.R >= 128 ? 2 : 0)
                    + (color.G >= 128 ? 4 : 0)
                    + (color.B >= 128 ? 1 : 0);
            return (ExColor)col;
        }

        // -------------------------------------------------------------
        // 透明パターンで塗りつぶす
        // -------------------------------------------------------------
        public static void FillTransparent(Graphics g, Rectangle pos)
        {
            g.InterpolationMode = InterpolationMode.NearestNeighbor;
            for (int y = 0; y < (pos.Height + 15) / 16; ++y)
            {
                for (int x = 0; x < (pos.Width + 15) / 16; ++x)
                {
                    int posX = pos.X + x * 16;
                    int posY = pos.Y + y * 16;
                    g.FillRectangle(Brushes.White, posX,     posY,     8, 8);
                    g.FillRectangle(brGray,        posX + 8, posY,     8, 8);
                    g.FillRectangle(brGray,        posX,     posY + 8, 8, 8);
                    g.FillRectangle(Brushes.White, posX + 8, posY + 8, 8, 8);
                }
            }
        }

        // -------------------------------------------------------------
        // 色変更用のイメージアトリビュートを取得する
        // -------------------------------------------------------------
        public static ImageAttributes GetImageAtribute(ExColor c)
        {
            return GetImageAtribute(GetColor(c));
        }
        public static ImageAttributes GetImageAtribute(Color c)
        {
            ColorMap[] cms = new ColorMap[]
            {
                new ColorMap(){ OldColor = Color.White, NewColor = c },
            };
            ImageAttributes ia = new ImageAttributes();
            ia.SetRemapTable(cms);
            return ia;
        }

        // -------------------------------------------------------------
        // 色変更反転用のイメージアトリビュートを取得する
        // -------------------------------------------------------------
        public ImageAttributes GetReverseAtribute(Color c)
        {
            ColorMap[] cms = new ColorMap[]
            {
                new ColorMap(){ OldColor = Color.White, NewColor = Color.Black },
                new ColorMap(){ OldColor = Color.Black, NewColor = c },
                new ColorMap(){ OldColor = Color.Transparent, NewColor = c },
            };
            ImageAttributes ia = new ImageAttributes();
            ia.SetRemapTable(cms);
            return ia;
        }

        // -------------------------------------------------------------
        // キャラコードから表示用の文字画像を取得する
        // -------------------------------------------------------------
        private Bitmap GetCharaPicture(int code)
        {
            using (Bitmap bmp = new Bitmap(CHIP_SIZE, CHIP_SIZE))
            using (Graphics g = Graphics.FromImage(bmp))
            {
                g.InterpolationMode = InterpolationMode.NearestNeighbor;
                Rectangle rcSrc = new Rectangle(code % 16 * CHIP_SIZE, code / 16 * CHIP_SIZE, CHIP_SIZE, CHIP_SIZE);
                Rectangle rcDest = new Rectangle(0, 0, CHIP_SIZE, CHIP_SIZE);
                g.DrawImage(pc80Font, rcDest, rcSrc, GraphicsUnit.Pixel);
                return (Bitmap)bmp.Clone();
            }
        }

        // -------------------------------------------------------------
        // 選択されているキャラをステータスバーに表示する
        // -------------------------------------------------------------
        private void DrawStatusCharactor(PictureBox pic, Chara chr)
        {
            if (pic.Image == null) return;
            using (Graphics g = Graphics.FromImage(pic.Image))
            {
                // 背景を塗りつぶす
                Rectangle rcDest = new Rectangle(2, 2, CHIP_SIZE * 2 + 1, CHIP_SIZE * 2 + 1);
                if (chr.code == 0)
                {
                    FillTransparent(g, rcDest);
                }
                else
                {
                    g.Clear((int)chr.color > 2 ? Color.Black : Color.White);
                }

                // 選択色で枠を描く
                Color color = GetColor(chr.color);
                g.DrawRectangle(new Pen(color, 2), 1, 1, SELICON_WIDTH - 2, SELICON_HEIGHT - 2);

                // 指定キャラ文字を表示する
                ImageAttributes ia = GetImageAtribute(color);
                g.DrawImage(
                    GetCharaPicture(chr.code),
                    rcDest, 0, 0, CHIP_SIZE, CHIP_SIZE,
                    GraphicsUnit.Pixel, ia
                );
            }
        }

        // -------------------------------------------------------------
        // キャラ文字編集サイズを取得する
        // -------------------------------------------------------------
        private Size GetChrSize(Layer layer = null)
        {
            if (layer == null)
            {
                if (lsCharas.Count <= 0) return new Size();
                layer = lsCharas[0];
            }
            return new Size(layer.chrTable.GetLength(ELEMENT_WIDTH), layer.chrTable.GetLength(ELEMENT_HEIGHT));
        }

        // -------------------------------------------------------------
        // 指定位置にキャラ文字を描画する
        // -------------------------------------------------------------
        private void DrawCharaChip(Graphics g, Chara chr, Rectangle rc)
        {
            using (Bitmap bmChr = GetCharaPicture(chr.code))
            using (Graphics o = Graphics.FromImage(bmChr))
            {
                // 特殊アトリビュート処理
                Color color = GetColor(chr.color);
                g.SetClip(rc);
                if ((chr.atrb & ATB_REVERSE) == ATB_REVERSE)
                {
                    g.Clear(color);
                    color = Color.Black;
                }
                else if (chr.code == 0)
                {
                    g.Clear(Color.Transparent);
                }
                else
                {
                    g.Clear(Color.Black);
                }
                if ((chr.atrb & ATB_OVERLINE) == ATB_OVERLINE)
                {
                    o.DrawLine(penWhite, 0, 0, CHIP_SIZE, 0);
                }
                if ((chr.atrb & ATB_UNDERLINE) == ATB_UNDERLINE)
                {
                    o.DrawLine(penWhite, 0, CHIP_SIZE - 1, CHIP_SIZE, CHIP_SIZE - 1);
                }
                ImageAttributes ia = (chr.atrb & ATB_REVERSE) != ATB_REVERSE
                    ? GetImageAtribute(color)
                    : GetReverseAtribute(color);

                // 描画する
                g.DrawImage(bmChr, rc, 0, 0, CHIP_SIZE, CHIP_SIZE, GraphicsUnit.Pixel, ia);
            }

        }

        // -------------------------------------------------------------
        // 指定位置にキャラ文字を設定する
        // -------------------------------------------------------------
        private void SetCharaTable(int num, Chara chr, Point pos)
        {
            // エラーチェック
            if (lsCharas.Count < num) return;
            Size sz = GetChrSize();
            if (pos.X < 0 || pos.X >= sz.Width) return;
            if (pos.Y < 0 || pos.Y >= sz.Height) return;

            // 配置する
            SetUndo();
            lsCharas[num].chrTable[pos.X, pos.Y] = chr;

            // イメージを更新する
            if (lsCharas[num].image == null) return;
            using (Graphics g = Graphics.FromImage(lsCharas[num].image))
            {
                // キャラ文字を表示する
                Rectangle rcDest = new Rectangle(pos.X * CHIP_SIZE, pos.Y * CHIP_SIZE, CHIP_SIZE, CHIP_SIZE);
                DrawCharaChip(g, chr, rcDest);
            }
            // サムネイルを更新する
            if (lsCharas[num].thumbnail != null) lsCharas[num].thumbnail.Dispose();
            Size szThumb = GetThumbSize();
            lsCharas[num].thumbnail = lsCharas[num].image.GetThumbnailImage(szThumb.Width, szThumb.Height, null, IntPtr.Zero);
            lvLayer.Refresh();
        }

        // -------------------------------------------------------------
        // 指定のキャラ文字テーブルを画像化する
        // -------------------------------------------------------------
        private void CreateLayerPicture(Layer layer)
        {
            Size sz = GetChrSize(layer);
            layer.image.Dispose();
            layer.image = new Bitmap(sz.Width * CHIP_SIZE, sz.Height * CHIP_SIZE);
            using (Graphics g = Graphics.FromImage(layer.image))
            {
                // キャラ文字を表示する
                Rectangle rcDest = new Rectangle(0, 0, CHIP_SIZE, CHIP_SIZE);
                for (int y = 0; y < sz.Height; ++y)
                {
                    rcDest.Y = y * CHIP_SIZE;
                    for (int x = 0; x < sz.Width; ++x)
                    {
                        Chara chr = layer.chrTable[x, y];
                        rcDest.X = x * CHIP_SIZE;
                        DrawCharaChip(g, chr, rcDest);
                    }
                }
            }

            // サムネイルを更新する
            if (layer.thumbnail != null) layer.thumbnail.Dispose();
            Size szThumb = GetThumbSize();
            layer.thumbnail = layer.image.GetThumbnailImage(szThumb.Width, szThumb.Height, null, IntPtr.Zero);
        }

        // -------------------------------------------------------------
        // リストビューを更新する
        // -------------------------------------------------------------
        private void UpdateLayer()
        {
            if (lsCharas.Count <= 0)
            {
                lvLayer.Clear();
                imgListLayer.Images.Clear();
                return;
            }
            CheangeScreenSize();
            lvLayer.BeginUpdate();
            lvLayer.Clear();
            imgListLayer.Images.Clear();
            Size szThumb = GetThumbSize();
            imgListLayer.ImageSize = new Size(szThumb.Width + 22, szThumb.Height + 12);
            for (int i = 0; i < lsCharas.Count; ++i)
            {
                lvLayer.Items.Add(lsCharas[i].label, i);
                lvLayer.Items[i].Checked = lsCharas[i].isShow;
                lvLayer.Items[i].Name = i.ToString();
                lvLayer.Items[i].Selected = (i == editLayer);
            }
            lvLayer.EnsureVisible(editLayer);
            lvLayer.EndUpdate();
        }

        // -------------------------------------------------------------
        // クライアント座標から配列座標を取得する
        // -------------------------------------------------------------
        private Point GetMapPos(Point pnt)
        {
            return GetMapPos(pnt.X, pnt.Y);
        }
        private Point GetMapPos(int x, int y)
        {
            if (lsCharas.Count <= 0) return new Point();
            Point pos = new Point();
            Size sz = GetChrSize();
            pos.X = x * sz.Width / picScreen.Width;
            pos.Y = y * sz.Height / picScreen.Height;
            return pos;
        }

        // -------------------------------------------------------------
        // 配列座標からクライアント座標を取得する
        // -------------------------------------------------------------
        private Point GetScreenPos(Point pnt)
        {
            return GetScreenPos(pnt.X, pnt.Y);
        }
        private Point GetScreenPos(int x, int y)
        {
            return new Point(GetScreenPosX(x), GetScreenPosY(y));
        }

        // -------------------------------------------------------------
        // X座標からクライアント座標を取得する
        // -------------------------------------------------------------
        private int GetScreenPosX(int x)
        {
            if (lsCharas.Count <= 0) return -1;
            return (int)(picScreen.Width * x / lsCharas[0].chrTable.GetLength(ELEMENT_WIDTH) + 0.5f);
        }

        // -------------------------------------------------------------
        // Y座標から実座標を取得する
        // -------------------------------------------------------------
        private int GetScreenPosY(int y)
        {
            if (lsCharas.Count <= 0) return -1;
            return (int)(picScreen.Height * y / lsCharas[0].chrTable.GetLength(ELEMENT_HEIGHT) + 0.5f);
        }

        // -------------------------------------------------------------
        // クライアント座標からキャラ文字を取得する
        // -------------------------------------------------------------
        private Chara GetMapChara(Point pnt)
        {
            return GetMapChara(pnt.X, pnt.Y);
        }
        private Chara GetMapChara(int x, int y)
        {
            if (lsCharas.Count <= 0) return new Chara(-1);
            Point pos = GetMapPos(x, y);
            return lsCharas[editLayer].chrTable[pos.X, pos.Y];
        }

        // -------------------------------------------------------------
        // 表示状態をひとつの配列にまとめる
        // -------------------------------------------------------------
        private Chara[,] GetJoinLayer(bool show = true)
        {
            Size sz = GetChrSize();
            if (sz.IsEmpty) return new Chara[0, 0];
            Chara[,] charas = new Chara[sz.Width, sz.Height];

            for (int i = lsCharas.Count - 1; i >= 0; --i)
            {
                if (show && !lsCharas[i].isShow) continue;
                sz = GetChrSize(lsCharas[i]);
                for (int y = 0; y < sz.Height; ++y)
                {
                    for (int x = 0; x < sz.Width; ++x)
                    {
                        if (lsCharas[i].chrTable[x, y].code != 0)
                        {
                            charas[x, y] = lsCharas[i].chrTable[x, y];
                        }
                    }
                }
            }
            return charas;
        }

        // -------------------------------------------------------------
        // グリッドを表示する
        // -------------------------------------------------------------
        private void DrawGrid(Graphics e)
        {
            if (zoomCnt < 2) return;
            Size sz = GetChrSize();
            Size sizeZoom = new Size(
                width: (int)(sz.Width * CHIP_SIZE * zoomRates[zoomCnt] + 0.5f),
                height: (int)(sz.Height * CHIP_SIZE * zoomRates[zoomCnt] + 0.5f)
            );
            for (int y = 0; y <= sz.Height; ++y)
            {
                float posY = GetScreenPosY(y);
                e.DrawLine(penGrid, 0, posY, sizeZoom.Width, posY);
            }
            for (int x = 0; x <= sz.Width; ++x)
            {
                float posX = GetScreenPosX(x);
                e.DrawLine(penGrid, posX, 0, posX, sizeZoom.Height);
            }

        }

        // -------------------------------------------------------------
        // アトリビュートイメージを表示する
        // -------------------------------------------------------------
        private void DrawAtrbImage(Graphics e = null)
        {
            Size sz = GetChrSize();
            int szUnit = (int)((float)picScreen.Width / lsCharas[0].chrTable.GetLength(ELEMENT_WIDTH) + 0.5f);
            Chara[,] charas = GetJoinLayer();
            lsAtrb.Clear();
            for (int y = 0; y < sz.Height; ++y)
            {
                // 最初に出てくる色を探す
                ExColor color = ExColor.White;
                for (int x = 0; x < TEXT_WIDTH; ++x)
                {
                    if (charas[x, y].code == 0) continue;
                    color = charas[x, y].color;
                    break;
                }

                int atrb = ATB_NORMAL;
                float posY = GetScreenPosY(y);
                int num = ViewBlackToolStripMenuItem.Checked ? 1 : 0;   // アトリビュート変更回数
                for (int x = 0; x < sz.Width; ++x)
                {
                    Chara chr = charas[x, y];
                    int beforeAtrb = atrb;
                    atrb = chr.atrb;

                    if (ViewBlackToolStripMenuItem.Checked)
                    {
                        // 背景が黒表示なら透明を変化点と数えない
                        if (chr.code != 0 && chr.color != ExColor.Transparent)
                        {
                            if (color != chr.color) ++num;
                            color = chr.color;
                        }
                    }
                    else
                    {
                        // 背景が透明なら透明を変化点とする
                        ExColor nowColor = chr.code != 0 ? chr.color : ExColor.Transparent;
                        if (color != nowColor) ++num;
                        color = nowColor;
                    }

                    float posX = GetScreenPosX(x);
                    if (color != ExColor.Transparent && e != null)
                    {
                        e.FillRectangle(brColors[(int)color], posX, posY, szUnit, szUnit);
                    }
                    if ((atrb != ATB_NORMAL || beforeAtrb != ATB_NORMAL) && e != null)
                    {
                        e.DrawRectangle(new Pen(GetColor((ExColor)((timerCounter % 2) + 1)), 2), posX, posY, szUnit, szUnit);
                    }
                    if (atrb != beforeAtrb) ++num;
                }
                lsAtrb.Add(num);
            }

            // 表示ラインの確認
            if (e == null) return;
            if (atrbPrnLine < 0 || atrbPrnLine >= lsAtrb.Count) atrbPrnLine = lsAtrb.IndexOf(lsAtrb.Max());
            int cntAtrb = lsAtrb[atrbPrnLine];

            if (cntAtrb > 15) lbInformation.BackColor = Color.Red;
            else lbInformation.BackColor = alartColors[cntAtrb];
            if (cntAtrb > 20)
            {
                lbInformation.ForeColor = timerCounter % 10 < 5 ? Color.Red : Color.White;
                picAtrbMarker.Image = timerCounter % 10 < 5 ? Properties.Resources.markRed : Properties.Resources.markOut;
            }
            else
            {
                lbInformation.ForeColor = cntAtrb >= 13 ? Color.White : Color.Black;
                if (cntAtrb < 10)      picAtrbMarker.Image = Properties.Resources.markGreen;
                else if (cntAtrb < 15) picAtrbMarker.Image = Properties.Resources.markYellow;
                else                   picAtrbMarker.Image = Properties.Resources.markRed;
            }
            lbInformation.Text = " Line = " + atrbPrnLine + ", Max Atrb : " + cntAtrb + " ";
            lbInformation.BorderStyle = BorderStyle.Fixed3D;
            picAtrbMarker.Location = new Point(picScreen.Location.X - 16, picScreen.Location.Y + GetScreenPosY(atrbPrnLine) + (szUnit - 16) / 2);
            picAtrbMarker.Size = new Size(16, 16);
        }

        // -------------------------------------------------------------
        // マウスカーソルを表示する
        // -------------------------------------------------------------
        private void DrawMouseCursor(Graphics e)
        {
            // マウスカーソル表示
            Size szCurosr = new Size((int)(CHIP_SIZE * zoomRates[zoomCnt] + 0.5f), (int)(CHIP_SIZE * zoomRates[zoomCnt] + 0.5f));
            Point posMouse = GetScreenPos(GetMapPos((picScreen.PointToClient(Cursor.Position))));
            if (picScreen.ClientRectangle.Contains(posMouse))
            {
                Rectangle rc = new Rectangle(posMouse, szCurosr);
                e.DrawRectangle(new Pen(penColors[(timerCounter % 7) + 1], 2), rc);
            }
        }

        // -------------------------------------------------------------
        // 編集カーソルを表示する
        // -------------------------------------------------------------
        private void DrawEditCursor(Graphics e)
        {
            if (mode != Cmd.Text && !ViewEditCursorToolStripMenuItem.Checked) return;
            if (drawCursor || mode == Cmd.PenOn || mode == Cmd.SelectMove || mode == Cmd.MoveOn) return;
            Point pnt = GetScreenPos(posCursor);
            if (!picScreen.ClientRectangle.Contains(posCursor)) return;

            // 編集カーソル表示
            Rectangle rcSel = GetRectSelect();
            if (rcSel.IsEmpty || rcSel.Contains(pnt))
            {
                Size szCurosr = new Size((int)(CHIP_SIZE * zoomRates[zoomCnt] + 0.5f), (int)(CHIP_SIZE * zoomRates[zoomCnt] + 0.5f));
                Rectangle rc = new Rectangle(GetScreenPos(posCursor), szCurosr);
                e.FillRectangle(Brushes.Gray, rc);
            }
        }

        // -------------------------------------------------------------
        // 表示イメージを作成する
        // -------------------------------------------------------------
        private void DrawScreenImage(Graphics o)
        {
            o.InterpolationMode = o.InterpolationMode = InterpolationMode.NearestNeighbor;
            o.PixelOffsetMode = o.PixelOffsetMode = PixelOffsetMode.Half;

            // レイヤー描画を積み重ねる
            Size sz = GetChrSize();
            int scrWidth = sz.Width * CHIP_SIZE;
            int scrHeight = sz.Height * CHIP_SIZE;
            Rectangle rcLayer = new Rectangle(0, 0, scrWidth, scrHeight);
            for (int layer = lsCharas.Count - 1; layer >= 0; --layer)
            {
                if (!lsCharas[layer].isShow) continue;
                o.DrawImage(lsCharas[layer].image, rcLayer, 0, 0, scrWidth, scrHeight, GraphicsUnit.Pixel);
                if (moveBlock.layer == layer)
                {
                    // 移動ブロックを表示する
                    int x = moveBlock.range.X * CHIP_SIZE;
                    int y = moveBlock.range.Y * CHIP_SIZE;
                    int w = moveBlock.chrTable.GetLength(ELEMENT_WIDTH) * CHIP_SIZE;
                    int h = moveBlock.chrTable.GetLength(ELEMENT_HEIGHT) * CHIP_SIZE;
                    o.DrawImage(moveBlock.image, new Rectangle(x, y, w, h), 0, 0, w, h, GraphicsUnit.Pixel);
                }
            }
        }

        // -------------------------------------------------------------
        // 編集エリアが再描画される
        // -------------------------------------------------------------
        private void picScreen_Paint(object sender, PaintEventArgs e)
        {
            if (!enableDrawScreen) return;
            Size sz = GetChrSize();
            if (sz.IsEmpty || bmpScreen == null)
            {
                e.Graphics.Clear(Color.Gray);
                return;
            }

            using (Graphics o = Graphics.FromImage(bmpScreen))
            {
                e.Graphics.InterpolationMode = o.InterpolationMode = InterpolationMode.NearestNeighbor;
                e.Graphics.PixelOffsetMode = o.PixelOffsetMode = PixelOffsetMode.Half;

                // 透明パターンで塗りつぶす
                int scrWidth = sz.Width * CHIP_SIZE;
                int scrHeight = sz.Height * CHIP_SIZE;
                Rectangle rcDest = new Rectangle(0, 0, (int)(scrWidth * zoomRates[zoomCnt] + 0.5f), (int)(scrHeight * zoomRates[zoomCnt] + 0.5f));
                if (!ViewBlackToolStripMenuItem.Checked)
                {
                    FillTransparent(e.Graphics, rcDest);
                }

                // レイヤー描画を積み重ねる
                o.Clear(Color.Transparent);
                DrawScreenImage(o);

                // 編集エリアに描画する
                Rectangle rcLayer = new Rectangle(0, 0, scrWidth, scrHeight);
                e.Graphics.DrawImage(bmpScreen, rcDest, rcLayer, GraphicsUnit.Pixel);

                // その他表示
                if (ViewGridToolStripMenuItem.Checked) DrawGrid(e.Graphics);                            // グリッドを描画する
                if (ViewAtrbToolStripMenuItem.Checked && mode != Cmd.MoveOn) DrawAtrbImage(e.Graphics); // アトリビュート表示をする
                if (!rcSelect.IsEmpty)
                {
                    // 選択枠を描画する
                    e.Graphics.DrawRectangle(penSelects[0], GetRectSelect());
                    e.Graphics.DrawRectangle(penSelects[1], GetRectSelect());
                }
                DrawMouseCursor(e.Graphics);
                switch (mode)
                {
                    case Cmd.Pen:
                    case Cmd.PenOn:
                    case Cmd.Text:
                        // カーソルを表示する
                        DrawEditCursor(e.Graphics);
                        break;
                }
            }
        }

        // -------------------------------------------------------------
        // カーソルが選択範囲内か
        // -------------------------------------------------------------
        private bool IsCurosrSelectRange()
        {
            Rectangle rc = GetRectSelect();
            if (rc.IsEmpty) rc = picScreen.ClientRectangle;
            return rc.Contains(GetScreenPos(posCursor));
        }

        // -------------------------------------------------------------
        // 選択枠の範囲を取得する
        // -------------------------------------------------------------
        private Rectangle GetRectSelect()
        {
            if (rcSelect.IsEmpty) return new Rectangle();
            int posX = GetScreenPosX(rcSelect.X);
            int posY = GetScreenPosY(rcSelect.Y);
            Rectangle rc = new Rectangle
            {
                X = posX,
                Y = posY,
                Width = GetScreenPosX(rcSelect.Width + rcSelect.X) - posX,
                Height = GetScreenPosY(rcSelect.Height + rcSelect.Y) - posY
            };
            return rc;
        }

        // -------------------------------------------------------------
        // クライアントエリアのサイズが変更された
        // -------------------------------------------------------------
        private void splitMain_Panel1_SizeChanged(object sender, EventArgs e)
        {
            Point pos = new Point();
            pos.X = (splitMain.Panel1.ClientSize.Width - picScreen.Width) / 2;
            pos.Y = (splitMain.Panel1.ClientSize.Height - picScreen.Height) / 2;
            if (pos.X < 0) pos.X = 0;
            if (pos.Y < 0) pos.Y = 0;
            picScreen.Location = pos;
        }

        // -------------------------------------------------------------
        // サムネイルサイズを取得する
        // -------------------------------------------------------------
        public Size GetThumbSize()
        {
            Size sz = GetChrSize();
            return !sz.IsEmpty
                ? new Size(THUMBNAIL_WIDTH, sz.Height * THUMBNAIL_WIDTH / sz.Width)
                : new Size(THUMBNAIL_WIDTH, TEXT_HEIGHT * THUMBNAIL_WIDTH / TEXT_WIDTH);
        }

        // -------------------------------------------------------------
        // サムネイル背景を作成する
        // -------------------------------------------------------------
        private void CreatelayerThumbBack()
        {
            Size szThumb = GetThumbSize();
            if (layerThumbBack != null) layerThumbBack.Dispose();
            layerThumbBack = new Bitmap(szThumb.Width, szThumb.Height);
            using (Graphics g = Graphics.FromImage(layerThumbBack))
            {
                Rectangle rcDest = new Rectangle(0, 0, layerThumbBack.Width + 2, layerThumbBack.Height + 2);
                FillTransparent(g, rcDest);
            }
        }

        // -------------------------------------------------------------
        // 編集エリアのサイズを変更する
        // -------------------------------------------------------------
        private void CheangeScreenSize()
        {
            Size sizeLayer = GetChrSize();
            if (sizeLayer.IsEmpty) return;
            picScreen.Width = (int)(sizeLayer.Width * CHIP_SIZE * zoomRates[zoomCnt]);
            picScreen.Height = (int)(sizeLayer.Height * CHIP_SIZE * zoomRates[zoomCnt]);
            if (bmpScreen != null) bmpScreen.Dispose();

            bmpScreen = new Bitmap(sizeLayer.Width * CHIP_SIZE, sizeLayer.Height * CHIP_SIZE);
            CreatelayerThumbBack();
            zoomText.Text = string.Format("{0, -8}", zoomRates[zoomCnt].ToString("P1"));
            splitMain_Panel1_SizeChanged(null, null);
        }

        // -------------------------------------------------------------
        // 非表示警告を処理する
        // -------------------------------------------------------------
        private DialogResult CheckShowLayer()
        {
            if (lsCharas.Count <= 0 || lsCharas[editLayer].isShow) return DialogResult.None;
            return MessageBox.Show(
                "このレイヤーは非表示になっています。" + System.Environment.NewLine
                + "続行するには、レイヤーを表示する必要があります。" + System.Environment.NewLine
                + "レイヤーを表示させますか？",
                productTitle,
                MessageBoxButtons.YesNo,
                MessageBoxIcon.Information
            );
        }

        // -------------------------------------------------------------
        // キャラ文字を配置する
        // -------------------------------------------------------------
        private void OnClickPen()
        {
            if (lsCharas.Count <= 0) return;

            // そのレイヤーは見えているか
            switch (CheckShowLayer())
            {
                case DialogResult.Yes:
                    lvLayer.Items[editLayer].Checked = true;
                    return;     // ダイアログ選択で位置が変わるのでキャラ配置は一旦キャンセル
                case DialogResult.No:
                    return;
            }

            // 編集カーソルを使用しているか
            if (!ViewEditCursorToolStripMenuItem.Checked)
            {
                posCursor = GetMapPos((picScreen.PointToClient(Cursor.Position)));
            }

            // 選択範囲が無いまたは選択範囲内なのでキャラ文字を置く
                Chara chr = (Control.ModifierKeys & Keys.Shift) != Keys.Shift ? charaSelect : charaBack;
            if (!IsCurosrSelectRange()) return;
            if (lsCharas[editLayer].chrTable[posCursor.X, posCursor.Y].code != chr.code
            || lsCharas[editLayer].chrTable[posCursor.X, posCursor.Y].color != chr.color
            || lsCharas[editLayer].chrTable[posCursor.X, posCursor.Y].atrb != chr.atrb)
            {
                SetCharaTable(editLayer, chr, posCursor);
                atrbPrnLine = -1;
                picScreen.Invalidate();
            }
        }

        // -------------------------------------------------------------
        // 選択文字画像がマウスで押された
        // -------------------------------------------------------------
        void picSelect_MouseClick(object sender, EventArgs e)
        {
            OnChangeChara(true);
        }

        // -------------------------------------------------------------
        // バック文字画像がマウスで押された
        // -------------------------------------------------------------
        void picBack_MouseClick(object sender, EventArgs e)
        {
            OnChangeChara(false);
        }

        // -------------------------------------------------------------
        // キャラ文字を選択する
        // -------------------------------------------------------------
        private void OnChangeChara(bool front)
        {
            if ((Control.ModifierKeys & Keys.Alt) == Keys.Alt)
            {
                Chara chr = GetMapChara(picScreen.PointToClient(Cursor.Position));
                if (chr.code != -1)
                {
                    SetSelectChara(chr);
                }
            }
            else
            {
                // キャラ文字変更を開く
				Chara chr = new Chara(front ? charaSelect : charaBack);
                Size sz = GetChrSize();
                Point pnt = posCursor;
                if (pnt.X >= 0 && pnt.X < sz.Width && pnt.Y >= 0 && pnt.Y < sz.Height)
                {
                    pnt.X = GetScreenPosX(pnt.X);
                    pnt.Y = GetScreenPosY(pnt.Y);
                }
                else
                {
                    pnt.X = GetScreenPosX(sz.Width / 2);
                    pnt.Y = GetScreenPosY(sz.Height / 2);
                }
                pnt = picScreen.PointToScreen(pnt);

                SetCommand(Cmd.DlgPattern);
                dlgParts = new DlgParts(chr, pnt, front);
                dlgParts.ShowDialog();

                if (EditPenToolStripMenuItem.Checked) SetCommand(Cmd.Pen);
                else if (EditSelectToolStripMenuItem.Checked) SetCommand(Cmd.Select);
                else if (EditMoveToolStripMenuItem.Checked) SetCommand(Cmd.Move);
                else SetCommand(Cmd.Text);

                DrawStatusCharactor(picSelect, charaSelect);
                picSelect.Invalidate();
            }
        }

        // -------------------------------------------------------------
        // 選択範囲コマンドの待機にクリックされた
        // -------------------------------------------------------------
        private void OnClickSelectWait()
        {
            if (GetRectSelect().Width > 0 && IsCurosrSelectRange())
            {
                // 選択枠の範囲内なので選択枠移動に移行する
                SetCommand(Cmd.SelectMove);
                offsetSelect = GetMapPos(picScreen.PointToClient(Cursor.Position));
                offsetSelect.X = rcSelect.Location.X - offsetSelect.X;
                offsetSelect.Y = rcSelect.Location.Y - offsetSelect.Y;
            }
            else
            {
                // 選択枠設定に移行する
                SetCommand(Cmd.SelectOn);
                offsetSelect = GetMapPos(picScreen.PointToClient(Cursor.Position));
            }
        }

        // -------------------------------------------------------------
        // 指定範囲内を同一キャラで埋め尽くす
        // -------------------------------------------------------------
        private void FillCharaRect(Rectangle rc, Chara c)
        {
            SetUndo();

            Size sz = GetChrSize();
            for (int y = rc.Y; y < rc.Y + rc.Height; ++y)
            {
                if (y < 0 || y >= sz.Height) continue;
                for (int x = rc.X; x < rc.X + rc.Width; ++x)
                {
                    if (x < 0 || x >= sz.Width) continue;
                    lsCharas[editLayer].chrTable[x, y].atrb = c.atrb;
                    lsCharas[editLayer].chrTable[x, y].code = c.code;
                    lsCharas[editLayer].chrTable[x, y].color = c.color;
                }
            }
            atrbPrnLine = -1;
            CreateLayerPicture(lsCharas[editLayer]);
        }

        // -------------------------------------------------------------
        // ブロックにデータを取り込む
        // -------------------------------------------------------------
        private Block GetBlockData()
        {
            // 選択範囲を補正する
            AdjustSelectRange();
            Rectangle rc = rcSelect;
            if (rc.IsEmpty) rc = new Rectangle(new Point(), GetChrSize());

            // 選択範囲を全て取り込む
            Point pos = GetMapPos(picScreen.PointToClient(Cursor.Position));
            pos.X = rc.Location.X - pos.X;
            pos.Y = rc.Location.Y - pos.Y;
            Block block = new Block(editLayer, pos, rc.Size);
            block.range.Location = rc.Location;
            for (int y = 0; y < rc.Height; ++y)
            {
                for (int x = 0; x < rc.Width; ++x)
                {
                    block.chrTable[x, y] = new Chara(lsCharas[editLayer].chrTable[x + rc.X, y + rc.Y]);
                }
            }
            return block;
        }

        // -------------------------------------------------------------
        // キャラ文字を移動する
        // -------------------------------------------------------------
        private void OnClickMove()
        {
            // マウス座標を保存するため先にブロックデータを取り込む
            moveBlock = GetBlockData();

            // 見えているか確認
            if (DialogResult.No == CheckShowLayer()) return;
            lvLayer.Items[editLayer].Checked = true;

            // 移動状態に移行する
            SetCommand(Cmd.MoveOn);
            CreateLayerPicture(moveBlock);                          // イメージ画像を作成する
            FillCharaRect(moveBlock.range, new Chara());            // 移動元を消す
            rcSelect = new Rectangle();                             // 選択枠解除
        }

        // -------------------------------------------------------------
        // テキスト入力を開始する
        // -------------------------------------------------------------
        private void OnClickText()
        {
            // そのレイヤーは見えているか
            switch (CheckShowLayer())
            {
                case DialogResult.Yes:
                    lvLayer.Items[editLayer].Checked = true;
                    return;     // ダイアログ選択で位置が変わるのでテキスト入力は一旦キャンセル
                case DialogResult.No:
                    return;
            }

            // 現在のカーソル位置を入力開始位置とする
            if (!IsCurosrSelectRange())
            {
                MessageBox.Show(
                    text: "カーソルが選択範囲外です。" + Environment.NewLine
                        + "選択範囲内を指定するか、選択範囲を解除してください。",
                    caption: productTitle,
                    buttons: MessageBoxButtons.OK,
                    icon: MessageBoxIcon.Information
                );
                return;
            }

            // テキスト入力状態に移行する
            textBox.Location = new Point(-100000, -100000);
            textBox.Visible = true;
            textBox.Focus();
            SetCommand(Cmd.Text);
        }

        // -------------------------------------------------------------
        // 画面内に選択枠を補正する
        // -------------------------------------------------------------
        private void AdjustSelectRange()
        {
            Size sz = GetChrSize();
            if (rcSelect.X < 0)
            {
                rcSelect.Width += rcSelect.X;
                rcSelect.X = 0;
            }
            if (rcSelect.Y < 0)
            {
                rcSelect.Height += rcSelect.Y;
                rcSelect.Y = 0;
            }
            if (rcSelect.X + rcSelect.Width >= sz.Width)
            {
                rcSelect.Width = sz.Width - rcSelect.X;
            }
            if (rcSelect.Y + rcSelect.Height >= sz.Height)
            {
                rcSelect.Height = sz.Height - rcSelect.Y;
            }
        }

        // -------------------------------------------------------------
        // 選択範囲を設定している
        // -------------------------------------------------------------
        private void OnMoveSelectOn()
        {
            if (lsCharas.Count <= 0) return;
            Point endPos = GetMapPos(picScreen.PointToClient(Cursor.Position));
            int width = Math.Abs(endPos.X - offsetSelect.X);
            int height = Math.Abs(endPos.Y - offsetSelect.Y);
            rcSelect = new Rectangle(
                endPos.X < offsetSelect.X ? endPos.X : offsetSelect.X,
                endPos.Y < offsetSelect.Y ? endPos.Y : offsetSelect.Y,
                width, height
            );
            AdjustSelectRange();
        }

        // -------------------------------------------------------------
        // 選択範囲を移動している
        // -------------------------------------------------------------
        private void OnMoveSelectMove()
        {
            Point pnt = GetMapPos(picScreen.PointToClient(Cursor.Position));
            pnt.X += offsetSelect.X;
            pnt.Y += offsetSelect.Y;
            rcSelect.Location = pnt;
        }

        // -------------------------------------------------------------
        // 選択ブロックを移動している
        // -------------------------------------------------------------
        private void OnMoveBlock()
        {
            Point pnt = GetMapPos(picScreen.PointToClient(Cursor.Position));
            pnt.X += moveBlock.offsetPos.X;
            pnt.Y += moveBlock.offsetPos.Y;
            moveBlock.range.Location = pnt;
        }

        // -------------------------------------------------------------
        // ブロックデータをレイヤに転記する
        // -------------------------------------------------------------
        private void SetBlockData(Block block)
        {
            Size sz = GetChrSize();
            for (int y = 0; y < block.range.Height; ++y)
            {
                int posY = y + block.range.Y;
                if (posY >= sz.Height) break;
                if (posY < 0) continue;
                for (int x = 0; x < block.range.Width; ++x)
                {
                    int posX = x + block.range.X;
                    if (posX >= sz.Width) break;
                    if (posX < 0) continue;
                    if (block.chrTable[x, y].code == 0) continue;
                    lsCharas[editLayer].chrTable[posX, posY] = block.chrTable[x, y];
                }
            }
            atrbPrnLine = -1;
        }

        // -------------------------------------------------------------
        // 選択ブロックの位置が確定した
        // -------------------------------------------------------------
        private void OnMoveBlockEnd()
        {
            SetCommand(Cmd.Move);
            SetBlockData(moveBlock);
            rcSelect = new Rectangle();
            moveBlock.layer = -1;
            CreateLayerPicture(lsCharas[editLayer]);
            lvLayer.Refresh();
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // 編集エリアをマウスが移動している
        // -------------------------------------------------------------
        private void picScreen_MouseMove(object sender, MouseEventArgs e)
        {
            SetCursorIcon();
            switch (e.Button)
            {
                case MouseButtons.Left:
                    switch (mode)
                    {
                        case Cmd.Select:
                            SetCursorIcon();
                            break;
                        case Cmd.SelectOn:
                            OnMoveSelectOn();
                            break;
                        case Cmd.SelectMove:
                            OnMoveSelectMove();
                            break;
                        case Cmd.MoveOn:
                            OnMoveBlock();
                            break;
                        case Cmd.PenOn:
                            posCursor = GetMapPos(picScreen.PointToClient(Cursor.Position));
                            OnClickPen();
                            break;
                        default:
                            return;
                    }
                    break;
                case MouseButtons.Right:
                    break;
            }
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // 編集エリアがクリックされた
        // -------------------------------------------------------------
        private void picScreen_MouseDown(object sender, MouseEventArgs e)
        {
            picScreen.Focus();
            Point bakPos = posCursor;
            switch (e.Button)
            {
                case MouseButtons.Left:
                    posCursor = GetMapPos(picScreen.PointToClient(Cursor.Position));
                    switch (mode)
                    {
                        case Cmd.Pen:
                            SetCommand(Cmd.PenOn);
                            SetUndo();
                            enableUndo = false;
                            OnClickPen();
                            break;
                        case Cmd.Select:
                            OnClickSelectWait();
                            break;
                        case Cmd.Move:
                            OnClickMove();
                            break;
                        case Cmd.Text:
                            OnClickText();
                            break;
                    }
                    break;
                case MouseButtons.Right:
                    OnChangeChara(true);
                    break;
            }
        }

        // -------------------------------------------------------------
        // 編集エリアでのクリックが離された
        // -------------------------------------------------------------
        private void picScreen_MouseUp(object sender, MouseEventArgs e)
        {
            switch (mode)
            {
                case Cmd.PenOn:
                    SetCommand(Cmd.Pen);
                    enableUndo = true;
                    break;
                case Cmd.SelectOn:
                case Cmd.SelectMove:
                    SetCommand(Cmd.Select);
                    if (rcSelect.Width <= 0 || rcSelect.Height <= 0)
                    {
                        rcSelect = new Rectangle();
                    }
                    break;
                case Cmd.MoveOn:
                    OnMoveBlockEnd();
                    break;
            }
        }

        // -------------------------------------------------------------
        // ファイル／新規
        // -------------------------------------------------------------
        private void FileNewToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SetUndo();

            lvLayer.Items.Clear();
            imgListLayer.Images.Clear();
            lsCharas = new List<Layer>();
            lsCharas.Add(new Layer(TEXT_WIDTH, TEXT_HEIGHT));
            picScreen.Visible = true;
            SetEditLayer(0);
            CreateLayerPicture(lsCharas[0]);
            lsCharas[0].label = "Layer0";
            lvLayer.Refresh();
            UpdateLayer();
            Invalidate();
        }

        // -------------------------------------------------------------
        // 画像をキャラ文字配列に置き換える
        // -------------------------------------------------------------
        private Chara[,] GetBmpToCharaTable(Bitmap img)
        {
            Chara[,] chrTable = new Chara[img.Width, img.Height];
            for (int y = 0; y < img.Height; ++y)
            {
                for (int x = 0; x < img.Width; ++x)
                {
                    ExColor color = GetColor(img.GetPixel(x, y));
                    chrTable[x, y] = new Chara(color != ExColor.Transparent ? 0x87/* ■ */ : 0, color);
                }
            }
            return chrTable;
        }

        // -------------------------------------------------------------
        // データを読み込む
        // -------------------------------------------------------------
        private void FileOpen(string filename)
        {
            string ext = Path.GetExtension(filename);
            if (String.CompareOrdinal(ext, ".chr40") == 0)
            {
                SaveData data;
                using (FileStream fs = new FileStream(filename, FileMode.Open, FileAccess.Read))
                {
                    BinaryFormatter f = new BinaryFormatter();
                    try
                    {
                        data = (SaveData)f.Deserialize(fs);
                    }
                    catch (Exception e)
                    {
                        MessageBox.Show(
                            text: e.Message + Environment.NewLine + filename,
                            caption: "File Read Error",
                            buttons: MessageBoxButtons.OK,
                            icon: MessageBoxIcon.Error
                        );
                        return;
                    }
                }
                SetUndo();
                lsCharas = new List<Layer>(data.lsCharas);
                editLayer = data.editLayer;
            }
            else
            {
                Chara[,] chrTable;
                using (Bitmap img = (Bitmap)Image.FromFile(filename))
                {
                    chrTable = GetBmpToCharaTable(img);
                }
                FileNewToolStripMenuItem_Click(null, null);
                lsCharas[0].chrTable = chrTable;
                picScreen.Visible = true;
                SetEditLayer(0);
                CreateLayerPicture(lsCharas[0]);
                lvLayer.Refresh();
                UpdateLayer();
                Invalidate();

            }
            filePath = filename.Substring(0, filename.LastIndexOf(ext)) + ".chr40";
            alreadyEdited = false;


            UpdateLayer();
            lvLayer.Refresh();
            picScreen.Invalidate();
            PrintTitle();
        }

        // -------------------------------------------------------------
        // ファイル／開く
        // -------------------------------------------------------------
        private void FileOpneToolStripMenuItem_Click(object sender, EventArgs e)
        {
            //ダイアログを表示する
            OpenFileDialog ofd = new OpenFileDialog()
            {
                FileName = "chara.chr40",
                Title = "キャラクタファイルを指定してください",
                Filter = "キャラクタファイル(*.chr40)|*.chr40|画像ファイル(*.bmp;*.png;*.jpg;*.gif;*.tiff)|*.bmp;*.png;*.jpg;*.gif;*.tiff|すべてのファイル(*.*)|*.*",
                FilterIndex = 1
            };
            if (DialogResult.OK != ofd.ShowDialog()) return;

            FileOpen(ofd.FileName);
        }

        // -------------------------------------------------------------
        // データを現在の編集の新規レイヤーとして読み込む
        // -------------------------------------------------------------
        private void FileMerge(string filename)
        {
            string ext = Path.GetExtension(filename);
            Rectangle rcClip = new Rectangle(new Point(), GetChrSize());
            if (String.CompareOrdinal(ext, ".chr40") == 0)
            {
                SaveData data;
                using (FileStream fs = new FileStream(filename, FileMode.Open, FileAccess.Read))
                {
                    BinaryFormatter f = new BinaryFormatter();
                    try
                    {
                        data = (SaveData)f.Deserialize(fs);
                    }
                    catch (Exception e)
                    {
                        MessageBox.Show(
                            text: e.Message + Environment.NewLine + filename,
                            caption: "File Read Error",
                            buttons: MessageBoxButtons.OK,
                            icon: MessageBoxIcon.Error
                        );
                        return;
                    }
                }
                SetUndo();

                for (int i = data.lsCharas.Count - 1; i >= 0; --i)
                {
                    pasteBlock = new Block(editLayer, rcClip);
                    pasteBlock.chrTable = data.lsCharas[i].chrTable;

                    // レイヤを追加する
                    btnLayerAdd_Click(null, null);

                    // トリミングしてセンタリングする
                    Rectangle rc = GetLayerRectangle(pasteBlock);
                    pasteBlock.chrTable = TrimLayer(pasteBlock.chrTable, rc);
                    pasteBlock.range = rc;
                    Size sz = GetChrSize();
                    pasteBlock.range.X = (sz.Width - rc.Width) / 2;
                    pasteBlock.range.Y = (sz.Height - rc.Height) / 2;

                    // ペーストする
                    SetBlockData(pasteBlock);
                    CreateLayerPicture(lsCharas[editLayer]);
                }
                editLayer = 0;
            }
            else
            {
                Chara[,] chrTable;
                using (Bitmap img = (Bitmap)Image.FromFile(filename))
                {
                    chrTable = GetBmpToCharaTable(img);
                }

                pasteBlock = new Block(editLayer, rcClip);
                pasteBlock.chrTable = chrTable;

                // レイヤを追加する
                btnLayerAdd_Click(null, null);

                // トリミングしてセンタリングする
                Rectangle rc = GetLayerRectangle(pasteBlock);
                pasteBlock.chrTable = TrimLayer(pasteBlock.chrTable, rc);
                pasteBlock.range = rc;
                Size sz = GetChrSize();
                pasteBlock.range.X = (sz.Width - rc.Width) / 2;
                pasteBlock.range.Y = (sz.Height - rc.Height) / 2;

                // ペーストする
                SetBlockData(pasteBlock);
                CreateLayerPicture(lsCharas[editLayer]);
                editLayer = 0;
            }

            UpdateLayer();
            lvLayer.Refresh();
            picScreen.Invalidate();
            PrintTitle();

        }

        // -------------------------------------------------------------
        // ファイル／マージ
        // -------------------------------------------------------------
        private void FileMergeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            //ダイアログを表示する
            OpenFileDialog ofd = new OpenFileDialog()
            {
                FileName = "chara.chr40",
                Title = "キャラクタファイルを指定してください",
                Filter = "キャラクタファイル(*.chr40)|*.chr40|画像ファイル(*.bmp;*.png;*.jpg;*.gif;*.tiff)|*.bmp;*.png;*.jpg;*.gif;*.tiff|すべてのファイル(*.*)|*.*",
                FilterIndex = 1
            };
            if (DialogResult.OK != ofd.ShowDialog()) return;

            FileMerge(ofd.FileName);
        }

        // -------------------------------------------------------------
        // ファイル／上書き保存
        // -------------------------------------------------------------
        private void FileSaveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (lsCharas.Count <= 0) return;

            // ファイル名確認
            if (filePath == String.Empty)
            {
                FileSaveAsToolStripMenuItem_Click(null, null);
                return;
            }

            try
            {
                using (FileStream fs = new FileStream(filePath, FileMode.Create, FileAccess.Write))
                {
                    BinaryFormatter bf = new BinaryFormatter();
                    SaveData data = new SaveData()
                    {
                        lsCharas = new List<Layer>(lsCharas),
                        editLayer = editLayer
                    };
                    bf.Serialize(fs, data);
                    fs.Close();
                }
                alreadyEdited = false;
                PrintTitle();
            }
            catch
            {
                MessageBox.Show(
                    text: "ファイルが作成できません" + Environment.NewLine
                          + Path.GetFileNameWithoutExtension(filePath),
                    caption: MainForm.form.Text,
                    buttons: MessageBoxButtons.OK,
                    icon: MessageBoxIcon.Error
                );
            }
        }

        // -------------------------------------------------------------
        // ファイル／名前を付けて保存
        // -------------------------------------------------------------
        private void FileSaveAsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (lsCharas.Count <= 0) return;
            SaveFileDialog sfd = new SaveFileDialog()
            {
                FileName = filePath != String.Empty ? filePath : "chara.chr40",
                Title = "保存先のファイルを選択してください",
                Filter = "キャラクタファイル(*.chr40)|*.chr40|すべてのファイル(*.*)|*.*",
                FilterIndex = 1,
                InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.DesktopDirectory),
                RestoreDirectory = true,
            };
            if (sfd.ShowDialog() != DialogResult.OK) return;

            filePath = sfd.FileName;
            FileSaveToolStripMenuItem_Click(null, null);
        }

        // -------------------------------------------------------------
        // バイナリデータを作成する
        // -------------------------------------------------------------
        byte[] CreateBinaryFile(Chara[,] charas, bool sprite = false)
        {
            List<byte> lsBin = new List<byte>();
            int width = charas.GetLength(ELEMENT_WIDTH);
            int height = charas.GetLength(ELEMENT_HEIGHT);
            lsBin.Add((byte)height);                        // キャラの高さ
            lsBin.Add((byte)width);                         // キャラの幅
            for (int y = 0; y < height; ++y)
            {
                // アトリビュート出力
                byte beforeAtrb = 0;
                for (int x = 0; x < width; ++x)
                {
                    byte atrb = atrbs[(int)charas[x, y].color + 1];
                    atrb |= (byte)charas[x, y].atrb;
                    if (charas[x, y].code == 0)
                    {
						 atrb = sprite ? (byte)0 : beforeAtrb;
					}
                    lsBin.Add(atrb);                        // アトリビュート
                    beforeAtrb = atrb;
                }

                // データ出力
                for (int x = 0; x < width; ++x)
                {
                    lsBin.Add((byte)charas[x, y].code);     // コード
                }
            }
            return lsBin.ToArray();
        }

        // -------------------------------------------------------------
        // 見えているレイヤーを分解してバイナリ出力する
        // -------------------------------------------------------------
        byte[] CreateLayerSpriteBinary()
        {
            List<List<byte>> lsBins = new List<List<byte>>();
            for (int i = 0; i < lsCharas.Count; ++i)
            {
                if (!lsCharas[i].isShow) continue;

                // データを出力用に加工する
                Rectangle rcCanvas = GetLayerRectangle(lsCharas[i]);
                Chara[,] charas = TrimLayer(lsCharas[i].chrTable, rcCanvas);


                // エラーチェック
                if (IsExistAtrb(charas))
                {
                    if (DialogResult.OK != MessageBox.Show(
                        text: "特殊アトリビュートが含まれています。" + Environment.NewLine
                            + "正常な動作を保証できません。それでも出力しますか？",
                        caption: productTitle,
                        buttons: MessageBoxButtons.OKCancel,
                        icon: MessageBoxIcon.Warning
                    ))
                    {
                        return null;
                    }
                }

                byte[] datas = CreateBinaryFile(charas, true);
                if (datas == null) return null;
                lsBins.Add(new List<byte>(datas));
            }

            List<byte> lsBin = new List<byte>();
            int offset = lsBins.Count * 2;
            for (int i = 0; i < lsBins.Count; ++i)
            {
                lsBin.Add((byte)(offset % 256));
                lsBin.Add((byte)(offset / 256));
                offset += lsBins[i].Count;
            }
            for (int i = 0; i < lsBins.Count; ++i)
            {
                lsBin.AddRange(lsBins[i]);
            }
            return lsBin.ToArray();
        }

        // -------------------------------------------------------------
        // 特殊アトリビュートが含まれているか？
        // -------------------------------------------------------------
        bool IsExistAtrb(Chara[,] charas)
        {
            for (int x = 0; x < charas.GetLength(ELEMENT_WIDTH); ++x)
            {
                for (int y = 0; y < charas.GetLength(ELEMENT_HEIGHT); ++y)
                {
                    if (charas[x, y].atrb != ATB_NORMAL)
                    {
                        return true;
                    }
                }
            }
            return false;
        }

        // -------------------------------------------------------------
        // レイヤーを統合してバイナリ出力する
        // -------------------------------------------------------------
        byte[] CreateSpriteBinary()
        {
            // データを出力用に加工する
            Chara[,] charas = GetJoinLayer();                   // 全てのレイヤを合成する
            Rectangle rcCanvas = GetLayerFitRect();             // トリミングサイズを取得する
            charas = TrimLayer(charas, rcCanvas);               // データをトリミングする

            // エラーチェック
            if (IsExistAtrb(charas))
            {
                if (DialogResult.OK != MessageBox.Show(
                    text: "特殊アトリビュートが含まれています。" + Environment.NewLine
                        + "正常な動作を保証できません。それでも出力しますか？",
                    caption: productTitle,
                    buttons: MessageBoxButtons.OKCancel,
                    icon: MessageBoxIcon.Warning
                ))
                {
                    return null;
                }
            }

            // バイナリファイルを出力する
            return CreateBinaryFile(charas);
        }

        // -------------------------------------------------------------
        // 画面イメージでバイナリを作成する
        // -------------------------------------------------------------
        byte[] CreateScreenImageBinary()
        {
            Size sz = GetChrSize();
            if (sz.Width != TEXT_WIDTH || sz.Height != TEXT_HEIGHT)
            {
                MessageBox.Show("画面イメージを出力できるのは画面サイズが幅" + TEXT_WIDTH + ",高さ" + TEXT_HEIGHT + "だけです。", productTitle, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return null;
            }

            // 画面イメージを作成する
            Chara[,] charas = GetJoinLayer();
            List<byte> lsBin = new List<byte>();
            for (int y = 0; y < TEXT_HEIGHT; ++y)
            {
                // テキストエリアを出力する
                for (int x = 0; x < TEXT_WIDTH; ++x)
                {
                    byte code = (byte)charas[x, y].code;
                    lsBin.Add(code);
                    lsBin.Add(code);
                }

                // 最初に出てくる色を探す
                ExColor color = ExColor.White;
                for (int x = 0;  x < TEXT_WIDTH; ++x)
                {
                    if (charas[x, y].code == 0) continue;
                    color = charas[x, y].color;
                    break;
                }

                // アトリビュートエリアのデータを出力する
                int atrb = ATB_NORMAL;
                lsBin.Add((byte)0x00);
                lsBin.Add(pc8001atrbs[(int)color]);
                int atrbCnt = 19;
                for (int x = 1; x < TEXT_WIDTH; ++x)
                {
                    Chara chr = charas[x, y];
                    int beforeAtrb = atrb;
                    atrb = chr.atrb;
                    if (atrb != beforeAtrb)
                    {
                        // 特殊アトリビュートの変化点を出力する
                        lsBin.Add((byte)(x * 2 - 1));
                        lsBin.Add((byte)(
                              ((atrb & ATB_REVERSE)   == ATB_REVERSE   ? 0x04 : 0x00)
                            | ((atrb & ATB_OVERLINE)  == ATB_OVERLINE  ? 0x10 : 0x00)
                            | ((atrb & ATB_UNDERLINE) == ATB_UNDERLINE ? 0x20 : 0x00)
                        ));
                        --atrbCnt;
                        if (atrbCnt < 0)
                        {
                            MessageBox.Show(
                                text: "アトリビュート変化数が20を超えましたので、" + Environment.NewLine
                                    + "バイナリ変換が出来ませんでした。",
                                caption: productTitle,
                                buttons: MessageBoxButtons.OK,
                                icon: MessageBoxIcon.Error
                            );
                            return null;
                        }
                    }

                    if (chr.code != 0 && chr.color != ExColor.Transparent)
                    {
                        if (color != chr.color)
                        {
                            // 色の変化点を出力する
                            lsBin.Add((byte)(x * 2));
                            lsBin.Add(pc8001atrbs[(int)chr.color]);
                            --atrbCnt;
                            if (atrbCnt < 0)
                            {
                                MessageBox.Show(
                                    text: "アトリビュート変化数が20を超えましたので、" + Environment.NewLine
                                        + "バイナリ変換が出来ませんでした。",
                                    caption: productTitle,
                                    buttons: MessageBoxButtons.OK,
                                    icon: MessageBoxIcon.Error
                                );
                                return null;
                            }
                        }
                        color = chr.color;
                    }
                }

                // アトリビュートの余り分を白で埋め尽くす
                for (; atrbCnt > 0; --atrbCnt)
                {
                    lsBin.Add((byte)0x50);
                    lsBin.Add((byte)0xE8);
                }
            }

            // Exomizer があれば圧縮データに変換する
            string exoFile = Path.GetDirectoryName(Path.GetFullPath(Environment.GetCommandLineArgs()[0]));
            exoFile += "\\exomizer.exe";
            try
            {
                if (File.Exists(exoFile))
                {
                    string tmpFile1 = Path.GetTempFileName();
                    string tmpFile2 = Path.GetTempFileName();
                    using (FileStream fs = new FileStream(tmpFile1, FileMode.Create))
                    using (BinaryWriter bw = new BinaryWriter(fs))
                    {
                        bw.Write(lsBin.ToArray());
                    }

                    ProcessStartInfo psi = new ProcessStartInfo();
                    psi.FileName = exoFile;
                    psi.Arguments = "raw -P0 \"" + tmpFile1 + "\" -o\"" + tmpFile2 + "\"";
                    psi.WindowStyle = ProcessWindowStyle.Hidden;
                    Process p = Process.Start(psi);
                    p.WaitForExit();
                    File.Delete(tmpFile1);

                    lsBin.Clear();
                    FileInfo fi = new FileInfo(tmpFile2);
                    using (FileStream fs = new FileStream(tmpFile2, FileMode.Open))
                    using (BinaryReader br = new BinaryReader(fs))
                    {
                        lsBin.AddRange(br.ReadBytes((int)fi.Length));
                    }
                    File.Delete(tmpFile2);
                }
            }
            catch
            {
                MessageBox.Show(
                    text: "バイナリファイルを変換できませんでした。",
                    caption: productTitle,
                    buttons: MessageBoxButtons.OK,
                    icon: MessageBoxIcon.Error
                );
                return null;
            }
            return lsBin.ToArray();
        }

        // -------------------------------------------------------------
        // M88 + diskdrv形式のバイナリリストを作成する
        // -------------------------------------------------------------
        List<byte> CreateBinaryM88diskdrv(byte[] datas, UInt16 topAdrs, UInt16 endAdrs)
        {
            List<byte> lsOutput = new List<byte>();
            lsOutput.Add((byte)(topAdrs % 256));
            lsOutput.Add((byte)(topAdrs / 256));
            lsOutput.Add((byte)(endAdrs % 256));
            lsOutput.Add((byte)(endAdrs / 256));
            lsOutput.AddRange(datas);
            return lsOutput;
        }

        // -------------------------------------------------------------
        // M88 t88形式のバイナリリストを作成する
        // -------------------------------------------------------------
        List<byte> CreateBinaryM88t88(byte[] datas, UInt16 topAdrs, string gameFile)
        {
            M88t88 obj = new M88t88(datas, topAdrs, gameFile);
            List<byte> lsOutput = new List<byte>(obj.GetBinary());
            return lsOutput;
        }

        // -------------------------------------------------------------
        // j80 cmt形式のバイナリリストを作成する
        // -------------------------------------------------------------
        List<byte> CreateBinaryJ80cmt(byte[] datas, UInt16 topAdrs)
        {
            J80cmt obj = new J80cmt(datas, topAdrs);
            List<byte> lsOutput = new List<byte>(obj.GetBinary());
            return lsOutput;
        }

        // -------------------------------------------------------------
        // ファイル／バイナリ出力
        // -------------------------------------------------------------
        private void FileExportBinToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (lsCharas.Count <= 0) return;
            DrawAtrbImage(null);
            if (lsAtrb.Max() > 20)
            {
                MessageBox.Show("アトリビュート変化点が20を超えています。", productTitle, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }
            Size sz = GetChrSize();
            if (sz.Width > 255 || sz.Height > 255)
            {
                MessageBox.Show("サイズが255を超えているため実機で表示できません。", productTitle, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            CommonOpenFileDialog dialog = new CommonOpenFileDialog();
            dialog.InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.DesktopDirectory);
            dialog.DefaultFileName = filePath != String.Empty ? Path.GetFileNameWithoutExtension(filePath) + ".bin" : "chara.bin";

            CommonFileDialogComboBox cbOption = new CommonFileDialogComboBox();
            cbOption.Items.Add(new CommonFileDialogComboBoxItem("レイヤー統合出力"));
            cbOption.Items.Add(new CommonFileDialogComboBoxItem("レイヤー個別出力"));
            cbOption.Items.Add(new CommonFileDialogComboBoxItem("画面イメージ出力"));
            cbOption.SelectedIndex = 0;
            dialog.Controls.Add(cbOption);
            dialog.Filters.Add(new CommonFileDialogFilter("バイナリファイル(*.bin)", "*.bin"));
            dialog.Filters.Add(new CommonFileDialogFilter("すべてのファイル(*.*)", "*.*"));
            if (CommonFileDialogResult.Ok != dialog.ShowDialog()) return;

            // 出力データを作成する
            byte[] datas = null;
            switch (cbOption.SelectedIndex)
            {
                case 0:
                    // レイヤー統合バイナリ出力
                    datas = CreateSpriteBinary();
                    break;

                case 1:
                    // レイヤー分解バイナリ出力
                    datas = CreateLayerSpriteBinary();
                    break;

                case 2:
                    // 画面イメージバイナリ出力
                    datas = CreateScreenImageBinary();
                    break;
            }
            if (datas == null) return;

            string outputPath = dialog.FileName;
            if (Path.GetExtension(outputPath).Trim('.') == "")
            {
                outputPath = outputPath.Trim('.') + ".bin";
            }

            // ヘッダ確認
            DlgExpBinary dlgAddress = new DlgExpBinary();
            dlgAddress.sizeData = datas.Length;
            dlgAddress.numUDEndAddress.Value = 0xDFFF;
            dlgAddress.tbType.Text = cbOption.Items[cbOption.SelectedIndex].Text;
            dlgAddress.lbPath.Text = outputPath;
            if (DialogResult.OK != dlgAddress.ShowDialog()) return;

            List<byte> lsOutput = new List<byte>();
            UInt16 topAdrs = (UInt16)dlgAddress.numUDTopAddress.Value;
            UInt16 endAdrs = (UInt16)dlgAddress.numUDEndAddress.Value;
            switch (dlgAddress.cbOutputFormat.SelectedIndex)
            {
                case 1: lsOutput = CreateBinaryM88diskdrv(datas, topAdrs, endAdrs); break;
                case 2: lsOutput = CreateBinaryM88t88(datas, topAdrs, Path.GetFileName(outputPath)); break;
                case 3: lsOutput = CreateBinaryJ80cmt(datas, topAdrs); break;
                default: lsOutput.AddRange(datas); break;
            }

            try
            {
                using (FileStream fs = new FileStream(dlgAddress.lbPath.Text, FileMode.Create, FileAccess.Write))
                using (BinaryWriter bw = new BinaryWriter(fs))
                {
                    bw.Write(lsOutput.ToArray());
                }
            }
            catch
            {
                MessageBox.Show(
                    text: "バイナリファイルを出力できませんでした。",
                    caption: productTitle,
                    buttons: MessageBoxButtons.OK,
                    icon: MessageBoxIcon.Error
                );
            }
        }

        // -------------------------------------------------------------
        // 見えているレイヤーを分解してキャラ文字イメージ画像を出力する
        // -------------------------------------------------------------
        void OutputLayerCharaImage(string path)
        {
            for (int i = 0; i < lsCharas.Count; ++i)
            {
                if (!lsCharas[i].isShow) continue;

                // データを出力用に加工する
                Rectangle rcSrc = GetLayerRectangle(lsCharas[i]);
                rcSrc.X *= CHIP_SIZE;
                rcSrc.Y *= CHIP_SIZE;
                rcSrc.Width *= CHIP_SIZE;
                rcSrc.Height *= CHIP_SIZE;

                string filename = Path.GetDirectoryName(path) + "\\" + Path.GetFileNameWithoutExtension(path) + i.ToString("000")
                    + (lsCharas[i].label == String.Empty ? "Layer" : lsCharas[i].label)
                    + ".png";

                Rectangle rcDest = new Rectangle(0, 0, rcSrc.Width, rcSrc.Height);
                try
                {
                    using (Image img = new Bitmap(rcDest.Width, rcDest.Height))
                    using (Graphics o = Graphics.FromImage(img))
                    {
                        o.InterpolationMode = o.InterpolationMode = InterpolationMode.NearestNeighbor;
                        o.PixelOffsetMode = o.PixelOffsetMode = PixelOffsetMode.Half;

                        o.Clear(ViewBlackToolStripMenuItem.Checked ? Color.Black : Color.Transparent);
                        o.DrawImage(lsCharas[i].image, rcDest, rcSrc, GraphicsUnit.Pixel);

                        img.Save(filename, ImageFormat.Png);
                    }
                }
                catch
                {
                    MessageBox.Show(
                        text: "バイナリファイルを出力できませんでした。" + Environment.NewLine + filename,
                        caption: productTitle,
                        buttons: MessageBoxButtons.OK,
                        icon: MessageBoxIcon.Error
                    );
                }
            }
        }

        // -------------------------------------------------------------
        // レイヤーを統合してキャラ文字イメージ画像を出力する
        // -------------------------------------------------------------
        void OutputCharaImage(string path)
        {
            Size sz = GetChrSize();
            Rectangle rcDest = new Rectangle(0, 0, sz.Width * CHIP_SIZE, sz.Height * CHIP_SIZE);
            try
            {
                using (Image img = new Bitmap(rcDest.Width, rcDest.Height))
                using (Graphics o = Graphics.FromImage(img))
                {
                    o.InterpolationMode = o.InterpolationMode = InterpolationMode.NearestNeighbor;
                    o.PixelOffsetMode = o.PixelOffsetMode = PixelOffsetMode.Half;

                    o.Clear(ViewBlackToolStripMenuItem.Checked ? Color.Black : Color.Transparent);
                    DrawScreenImage(o);
                    img.Save(path, ImageFormat.Png);
                }
            }
            catch
            {
                MessageBox.Show(
                    text: "バイナリファイルを出力できませんでした。" + Environment.NewLine + path,
                    caption: productTitle,
                    buttons: MessageBoxButtons.OK,
                    icon: MessageBoxIcon.Error
                );
            }
        }

        // -------------------------------------------------------------
        // ファイル／イメージ画像出力／キャラ文字イメージ
        // -------------------------------------------------------------
        private void ExportCharaImageToolStripMenuItem_Click(object sender, EventArgs e)
        {
            CommonOpenFileDialog dialog = new CommonOpenFileDialog();
            dialog.InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.DesktopDirectory);
            dialog.DefaultFileName = filePath != String.Empty ? Path.GetFileNameWithoutExtension(filePath) + ".png" : "chara.png";

            CommonFileDialogComboBox cbOption = new CommonFileDialogComboBox();
            cbOption.Items.Add(new CommonFileDialogComboBoxItem("レイヤー統合出力"));
            cbOption.Items.Add(new CommonFileDialogComboBoxItem("レイヤー分解個別出力"));
            cbOption.SelectedIndex = 0;
            dialog.Controls.Add(cbOption);
            dialog.Filters.Add(new CommonFileDialogFilter("画像ファイル(*.png)", "*.png"));
            dialog.Filters.Add(new CommonFileDialogFilter("すべてのファイル(*.*)", "*.*"));

            if (CommonFileDialogResult.Ok != dialog.ShowDialog()) return;
            switch (cbOption.SelectedIndex)
            {
                case 0: OutputCharaImage(dialog.FileName); break;
                case 1: OutputLayerCharaImage(dialog.FileName); break;
            }
        }

        // -------------------------------------------------------------
        // 見えているレイヤーを分解してカラードットイメージを出力する
        // -------------------------------------------------------------
        void OutputLayerColorDotImage(string path)
        {
            for (int i = 0; i < lsCharas.Count; ++i)
            {
                if (!lsCharas[i].isShow) continue;

                string filename = Path.GetDirectoryName(path) + "\\" + Path.GetFileNameWithoutExtension(path) + i.ToString("000")
                    + (lsCharas[i].label == String.Empty ? "Layer" : lsCharas[i].label)
                    + ".png";

                // データを出力用に加工する
                Rectangle rcCanvas = GetLayerRectangle(lsCharas[i]);
                Chara[,] charas = TrimLayer(lsCharas[i].chrTable, rcCanvas);

                Rectangle rcDest = new Rectangle(0, 0, charas.GetLength(ELEMENT_WIDTH), charas.GetLength(ELEMENT_HEIGHT));
                using (Image img = new Bitmap(rcDest.Width, rcDest.Height))
                using (Graphics o = Graphics.FromImage(img))
                {
                    o.Clear(ViewBlackToolStripMenuItem.Checked ? Color.Black : Color.Transparent);
                    for (int y = 0; y < rcDest.Height; ++y)
                    {
                        for (int x = 0; x < rcDest.Width; ++x)
                        {
                            if (charas[x, y].code == 0) continue;
                            ((Bitmap)img).SetPixel(x, y, GetColor(charas[x, y].color));
                        }
                    }
                    img.Save(filename, ImageFormat.Png);
                }
            }
        }

        // -------------------------------------------------------------
        // レイヤーを統合してカラードットイメージを出力する
        // -------------------------------------------------------------
        void OutputColorDotImage(string path)
        {
            Chara[,] charas = GetJoinLayer();                   // 全てのレイヤを合成する
            Rectangle rcDest = new Rectangle(new Point(), GetChrSize());
            using (Image img = new Bitmap(rcDest.Width, rcDest.Height))
            using (Graphics o = Graphics.FromImage(img))
            {
                o.Clear(ViewBlackToolStripMenuItem.Checked ? Color.Black : Color.Transparent);
                for (int y = 0; y < rcDest.Height; ++y)
                {
                    for (int x = 0; x < rcDest.Width; ++x)
                    {
                        if (charas[x, y].code == 0) continue;
                        ((Bitmap)img).SetPixel(x, y, GetColor(charas[x, y].color));
                    }
                }
                img.Save(path, ImageFormat.Png);
            }
        }

        // -------------------------------------------------------------
        // ファイル／イメージ画像出力／カラードットイメージ
        // -------------------------------------------------------------
        private void ExportDotsImageToolStripMenuItem_Click(object sender, EventArgs e)
        {
            CommonOpenFileDialog dialog = new CommonOpenFileDialog();
            dialog.InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.DesktopDirectory);
            dialog.DefaultFileName = filePath != String.Empty ? Path.GetFileNameWithoutExtension(filePath) + ".png" : "chara.png";

            CommonFileDialogComboBox cbOption = new CommonFileDialogComboBox();
            cbOption.Items.Add(new CommonFileDialogComboBoxItem("レイヤー統合出力"));
            cbOption.Items.Add(new CommonFileDialogComboBoxItem("レイヤー分解個別出力"));
            cbOption.SelectedIndex = 0;
            dialog.Controls.Add(cbOption);
            dialog.Filters.Add(new CommonFileDialogFilter("画像ファイル(*.png)", "*.png"));
            dialog.Filters.Add(new CommonFileDialogFilter("すべてのファイル(*.*)", "*.*"));

            if (CommonFileDialogResult.Ok != dialog.ShowDialog()) return;
            switch (cbOption.SelectedIndex)
            {
                case 0: OutputColorDotImage(dialog.FileName); break;
                case 1: OutputLayerColorDotImage(dialog.FileName); break;
            }
        }

        // -------------------------------------------------------------
        // ファイル／終了
        // -------------------------------------------------------------
        private void FileExitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Close();
        }
        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (alreadyEdited)
            {
                e.Cancel = DialogResult.No == MessageBox.Show(
                    text: "保存していない編集データがあります。本当に終了しますか？",
                    caption: "終了確認",
                    buttons: MessageBoxButtons.YesNo,
                    icon: MessageBoxIcon.Question);
            }
        }

        // -------------------------------------------------------------
        // アンドゥ／リドゥを実際に動作させる
        // -------------------------------------------------------------
        private void UndoRedo()
        {
            enableDrawScreen = false;

            lsCharas = new List<Layer>(lsUndo[undoPos].lsCharas);
            posCursor = lsUndo[undoPos].posCursor;
            editLayer = lsUndo[undoPos].editLayer;
            filePath = lsUndo[undoPos].filename;
            foreach (Layer lay in lsCharas)
            {
                CreateLayerPicture(lay);
            }

            atrbPrnLine = -1;
            CheangeScreenSize();
            UpdateLayer();
            lvLayer.Refresh();
            PrintTitle();
            enableDrawScreen = true;
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // 編集／元に戻す
        // -------------------------------------------------------------
        private void EditUndoToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (!enableUndo || undoPos <= 0) return;
            if (undoPos == lsUndo.Count)
            {
                SetUndo();
                --undoPos;
            }
            --undoPos;
            UndoRedo();
        }

        // -------------------------------------------------------------
        // 編集／やり直し
        // -------------------------------------------------------------
        private void EditReDoToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (lsUndo.Count <= undoPos + 1) return;
            ++undoPos;
            UndoRedo();
        }

        // -------------------------------------------------------------
        // 編集／切り取り
        // -------------------------------------------------------------
        private void EditCutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Clipboard.Clear();
            pasteBlock = GetBlockData();                            // ブロックにデータを取り込む
            FillCharaRect(pasteBlock.range, new Chara());           // 移動元を消す
            lvLayer.Refresh();
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // 編集／コピー
        // -------------------------------------------------------------
        private void EditCopyToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Clipboard.Clear();
            pasteBlock = GetBlockData();                            // ブロックにデータを取り込む
        }

        // -------------------------------------------------------------
        // 指定レイヤーのデータ範囲を取得する
        // -------------------------------------------------------------
        private Rectangle GetLayerRectangle(Layer layer)
        {
            if (layer.chrTable.Length <= 0) return new Rectangle();
            int width = layer.chrTable.GetLength(ELEMENT_WIDTH);
            int height = layer.chrTable.GetLength(ELEMENT_HEIGHT);

            // 右端確認
            int[] rights = new int[height];
            rights = Enumerable.Repeat(-1, height).ToArray();
            for (int y = 0; y < height; ++y)
            {
                for (int x = width - 1; x >= 0; --x)
                {
                    if (layer.chrTable[x, y].code == 0) continue;
                    rights[y] = x;
                    break;
                }
            }
            Array.Sort(rights);
            Array.Reverse(rights);
            if (rights[0] < 0) return new Rectangle();

            // 左端確認
            int[] lefts = new int[height];
            lefts = Enumerable.Repeat(width, height).ToArray();
            for (int y = 0; y < height; ++y)
            {
                for (int x = 0; x < width; ++x)
                {
                    if (layer.chrTable[x, y].code == 0) continue;
                    lefts[y] = x;
                    break;
                }
            }
            Array.Sort(lefts);

            // 上端確認
            int[] ups = new int[width];
            ups = Enumerable.Repeat(height, width).ToArray();
            for (int x = 0; x < width; ++x)
            {
                for (int y = 0; y < height; ++y)
                {
                    if (layer.chrTable[x, y].code == 0) continue;
                    ups[x] = y;
                    break;
                }
            }
            Array.Sort(ups);

            // 下端確認
            int[] downs = new int[width];
            downs = Enumerable.Repeat(-1, width).ToArray();
            for (int x = 0; x < width; ++x)
            {
                for (int y = height - 1; y >= 0; --y)
                {
                    if (layer.chrTable[x, y].code == 0) continue;
                    downs[x] = y;
                    break;
                }
            }
            Array.Sort(downs);
            Array.Reverse(downs);

            return new Rectangle(lefts[0], ups[0], rights[0] - lefts[0] + 1, downs[0] - ups[0] + 1);
        }

        // -------------------------------------------------------------
        // 編集エリアのレイヤをトリミングする
        // -------------------------------------------------------------
        private void TrimEditLayer(Rectangle rc)
        {
            SetUndo();

            enableDrawScreen = false;
            for (int i = 0; i < lsCharas.Count; ++i)
            {
                lsCharas[i].chrTable = TrimLayer(lsCharas[i].chrTable, rc);
                CreateLayerPicture(lsCharas[i]);
            }
            CheangeScreenSize();
            UpdateLayer();
            lvLayer.Refresh();
            enableDrawScreen = true;
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // データを指定サイズにトリミングする
        // -------------------------------------------------------------
        private Chara[,] TrimLayer(Chara[,] src, Rectangle rc)
        {
            Chara nullChara = new Chara();
            Size szSrc = new Size();
            szSrc.Width = src.GetLength(ELEMENT_WIDTH);
            szSrc.Height = src.GetLength(ELEMENT_HEIGHT);

            Chara[,] charas = new Chara[rc.Width, rc.Height];
            for (int y = 0; y < rc.Height; ++y)
            {
                for (int x = 0; x < rc.Width; ++x)
                {
                    int posX = rc.X + x;
                    int posY = rc.Y + y;
                    charas[x, y] = posX >= 0 && posX < szSrc.Width && posY >= 0 && posY < szSrc.Height
                        ? src[posX, posY]
                        : nullChara;
                }
            }
            return charas;
        }

        // -------------------------------------------------------------
        // 編集／貼り付け
        // -------------------------------------------------------------
        private void EditPasteToolStripMenuItem_Click(object sender, EventArgs e)
        {
            // クリップボード確認
            if (Clipboard.ContainsImage())
            {
                Image img = Clipboard.GetImage();
                if (img != null)
                {
                    Rectangle rcClip = new Rectangle(new Point(), GetChrSize());
                    Point pos = GetMapPos(picScreen.PointToClient(Cursor.Position));
                    pos.X = rcClip.Location.X - pos.X;
                    pos.Y = rcClip.Location.Y - pos.Y;
                    pasteBlock = new Block(editLayer, pos, rcClip.Size);
                    pasteBlock.range.Location = rcClip.Location;
                    pasteBlock.chrTable = GetBmpToCharaTable((Bitmap)img);
                }
            }
            else
            {
                if (pasteBlock.layer < 0 || pasteBlock.chrTable.Length <= 0) return;
            }

            // レイヤを追加する
            btnLayerAdd_Click(null, null);

            // トリミングしてセンタリングする
            Rectangle rc = GetLayerRectangle(pasteBlock);
            pasteBlock.chrTable = TrimLayer(pasteBlock.chrTable, rc);
            pasteBlock.range = rc;
            Size sz = GetChrSize();
            pasteBlock.range.X = (sz.Width - rc.Width) / 2;
            pasteBlock.range.Y = (sz.Height - rc.Height) / 2;

            // ペーストする
            SetBlockData(pasteBlock);
            CreateLayerPicture(lsCharas[editLayer]);
            lvLayer.Refresh();
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // 編集／キャラ配置が選択された
        // -------------------------------------------------------------
        private void EditCharaSetToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SetCommand(Cmd.Pen);
        }

        // -------------------------------------------------------------
        // 編集／選択選択を選択した
        // -------------------------------------------------------------
        private void RectSelectToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SetCommand(Cmd.Select);
        }

        // -------------------------------------------------------------
        // 編集／全てを選択するを選択した
        // -------------------------------------------------------------
        private void EditSelectAllToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SetCommand(Cmd.Select);
            rcSelect.Location = new Point();
            rcSelect.Size = GetChrSize();
            SetCursorIcon();
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // 編集／選択を解除を選択した
        // -------------------------------------------------------------
        private void EditSelectDeleteToolStripMenuItem_Click(object sender, EventArgs e)
        {
            switch (mode)
            {
                case Cmd.Move: break;           // 移動を完了する
                case Cmd.Pen: break;            // ペンの時はそのまま
                default:                        // 選択範囲の時は
                    SetCommand(Cmd.Select);     // 最初に戻す
                    break;
            }
            rcSelect = new Rectangle();
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // 編集／塗りつぶすを選択した
        // -------------------------------------------------------------
        private void EditFillRectToolStripMenuItem_Click(object sender, EventArgs e)
        {
            // そのレイヤーは見えているか
            if (DialogResult.No == CheckShowLayer()) return;
            lvLayer.Items[editLayer].Checked = true;

            Rectangle rc = rcSelect.Width > 0
                ? rcSelect
                : new Rectangle(new Point(), GetChrSize());
            FillCharaRect(rc, sender != null ? charaSelect : new Chara());
            lvLayer.Refresh();
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // 編集／移動を選択した
        // -------------------------------------------------------------
        private void EditMoveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SetCommand(Cmd.Move);
        }

        // -------------------------------------------------------------
        // 編集／文字入力モードを選択した
        // -------------------------------------------------------------
        private void EditTextInputToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OnClickText();
        }

        // -------------------------------------------------------------
        // 現在の最適なキャンバスサイズを取得する
        // -------------------------------------------------------------
        private Rectangle GetLayerFitRect()
        {
            if (lsCharas.Count <= 0) return new Rectangle();

            // 編集サイズのキャンバスサイズを作成する
            Rectangle rcCanvas = GetLayerRectangle(lsCharas[0]);
            for (int i = 1; i < lsCharas.Count; ++i)
            {
                Rectangle rc = GetLayerRectangle(lsCharas[i]);
                if (rc.IsEmpty) continue;
                int right = rcCanvas.Right < rc.Right ? rc.Right : rcCanvas.Right;
                int bottom = rcCanvas.Bottom < rc.Bottom ? rc.Bottom : rcCanvas.Bottom;
                if (rcCanvas.X > rc.X) rcCanvas.X = rc.X;
                if (rcCanvas.Y > rc.Y) rcCanvas.Y = rc.Y;
                rcCanvas.Width = right - rcCanvas.X;
                rcCanvas.Height = bottom - rcCanvas.Y;
            }

            return rcCanvas;
        }

        // -------------------------------------------------------------
        // イメージメニューが開かれる
        // -------------------------------------------------------------
        private void ImageToolStripMenuItem_DropDownOpening(object sender, EventArgs e)
        {
            ImageClipToolStripMenuItem.Enabled = !GetRectSelect().IsEmpty;
            ImageConvWatchToolStripMenuItem.Enabled = ImageCanvasFitToolStripMenuItem.Enabled = !GetLayerFitRect().IsEmpty;
        }

        // -------------------------------------------------------------
        // イメージ／キャンバスサイズ
        // -------------------------------------------------------------
        private void ImageCanvasDlgToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (lsCharas.Count <= 0) return;

            // 新しいキャンバスサイズを取得する
            DlgCanvasSize dlg = new DlgCanvasSize();
            dlg = new DlgCanvasSize();
            DlgCanvasSize.sizeCanvas = GetChrSize();
            if (DialogResult.OK != dlg.ShowDialog()) return;

            Rectangle rcCanvas = new Rectangle(new Point(), GetChrSize());
            rcCanvas.X = (rcCanvas.Width - DlgCanvasSize.sizeCanvas.Width) / 2;
            rcCanvas.Y = (rcCanvas.Height - DlgCanvasSize.sizeCanvas.Height) / 2;
            rcCanvas.Width = DlgCanvasSize.sizeCanvas.Width;
            rcCanvas.Height = DlgCanvasSize.sizeCanvas.Height;

            TrimEditLayer(rcCanvas);
        }

        // -------------------------------------------------------------
        // イメージ／キャンバスサイズ／編集サイズに合わせる
        // -------------------------------------------------------------
        private void ImageCanvasFitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (lsCharas.Count <= 0) return;

            // 全てのレイヤをトリミングする
            Rectangle rcCanvas = GetLayerFitRect();
            if (rcCanvas.IsEmpty) return;
            TrimEditLayer(rcCanvas);
        }

        // -------------------------------------------------------------
        // イメージ／切り抜きが選択された
        // -------------------------------------------------------------
        private void ImageClipToolStripMenuItem_Click(object sender, EventArgs e)
        {
            TrimEditLayer(rcSelect);
            rcSelect = new Rectangle();
            mode = Cmd.Pen;
            PrintCommand();
        }

        // -------------------------------------------------------------
        // イメージ／見たまま画像変換が選択された
        // -------------------------------------------------------------
        private void ImageConvWatchToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (lsCharas.Count <= 0) return;
            Size sizeCanvas = GetChrSize();
            if (sizeCanvas.IsEmpty) return;

            sizeCanvas.Width *= CHIP_SIZE;
            sizeCanvas.Height *= CHIP_SIZE;
            int total = sizeCanvas.Width * sizeCanvas.Height * lsCharas.Count;
            if (sizeCanvas.Width >= 512 || sizeCanvas.Height >= 512 || total >= 32768)
            {
                if (DialogResult.Yes != MessageBox.Show(
                    text: "変換後の編集サイズが大きすぎます。" + Environment.NewLine
                        + "Width = " + sizeCanvas.Width + ", Height = " + sizeCanvas.Height + ", Size = " + total + Environment.NewLine
                        + Environment.NewLine
                        + "画像変換を強行しますか？",
                    caption: "見たまま画像変換",
                    buttons: MessageBoxButtons.YesNo,
                    icon: MessageBoxIcon.Warning
                ))
                {
                    return;
                }
            }

            SetUndo();
            Cursor = picScreen.Cursor = Cursors.WaitCursor;
            for (int i = 0; i < lsCharas.Count; ++i)
            {
                lsCharas[i].chrTable = GetBmpToCharaTable((Bitmap)lsCharas[i].image);
                CreateLayerPicture(lsCharas[i]);
            }
            Cursor = Cursors.Default;
            zoomCnt = 0;
            SetCursorIcon();

            lvLayer.Refresh();
            UpdateLayer();
            Invalidate();

        }

        // -------------------------------------------------------------
        // 表示／ズームイン
        // -------------------------------------------------------------
        private void ViewZoomInToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (zoomRates[zoomCnt] >= 10.0f) return;
            ++zoomCnt;
            CheangeScreenSize();
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // 表示／ズームアウト
        // -------------------------------------------------------------
        private void ViewZoomOutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (zoomRates[zoomCnt] <= 0.1f) return;
            --zoomCnt;
            CheangeScreenSize();
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // 表示／画面サイズに合わせる
        // -------------------------------------------------------------
        private void ViewZoomFitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Size sizeClient = splitMain.Panel1.Size;
            for (zoomCnt = zoomRates.Count() - 1; zoomCnt > 0; --zoomCnt)
            {
                Size sizePic = GetChrSize();
                sizePic.Width = (int)(sizePic.Width * CHIP_SIZE * zoomRates[zoomCnt]);
                sizePic.Height = (int)(sizePic.Height * CHIP_SIZE * zoomRates[zoomCnt]);
                if (sizeClient.Width >= sizePic.Width && sizeClient.Height >= sizePic.Height) break;
            }
            CheangeScreenSize();
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // 表示／ピクセル等価
        // -------------------------------------------------------------
        private void PixelEquToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (zoomRates[zoomCnt].Equals(1.0f)) return;
            zoomCnt = 3;
            CheangeScreenSize();
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // 表示／編集カーソル
        // -------------------------------------------------------------
        private void ViewEditCursorToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ViewEditCursorToolStripMenuItem.Checked = !ViewEditCursorToolStripMenuItem.Checked;
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // 表示／グリッド
        // -------------------------------------------------------------
        private void ViewGridToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ViewGridToolStripMenuItem.Checked = !ViewGridToolStripMenuItem.Checked;
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // 表示／アトリビュート
        // -------------------------------------------------------------
        private void ViewAtrbToolStripMenuItem_Click(object sender, EventArgs e)
        {
            picAtrbMarker.Visible
                = lbInformation.Visible
                = ViewAtrbToolStripMenuItem.Checked = !ViewAtrbToolStripMenuItem.Checked;
            if (!lbInformation.Visible) PrintCommand();
            atrbPrnLine = -1;
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // 表示／背景を黒色にする
        // -------------------------------------------------------------
        private void ViewBlackToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ViewBlackToolStripMenuItem.Checked = !ViewBlackToolStripMenuItem.Checked;
            atrbPrnLine = -1;
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // ヘルプ／バージョン番号
        // -------------------------------------------------------------
        private void HelpAboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            AboutBox dlg = new AboutBox();
            dlg.ShowDialog();
        }

        // -------------------------------------------------------------
        // レイヤの追加ボタンを押した
        // -------------------------------------------------------------
        private void btnLayerAdd_Click(object sender, EventArgs e)
        {
            SetUndo();

            Size szCanvas = GetChrSize();
            if (szCanvas.IsEmpty) szCanvas = new Size(TEXT_WIDTH, TEXT_HEIGHT);
            lsCharas.Insert(0, new Layer(szCanvas));
            lsCharas[0].label = "Layer" + (lsCharas.Count - 1);
            SetEditLayer(0);
            picScreen.Visible = true;
            CreateLayerPicture(lsCharas[0]);
            UpdateLayer();
        }

        // -------------------------------------------------------------
        // 選択レイヤの表示優先度を上げるボタンを押した
        // -------------------------------------------------------------
        private void btnLayerUp_Click(object sender, EventArgs e)
        {
            SetUndo();

            if (editLayer == 0 || lsCharas.Count <= 1) return;
            Layer tmp = new Layer(lsCharas[editLayer]);
            lsCharas[editLayer] = lsCharas[editLayer - 1];
            lsCharas[editLayer - 1] = tmp;
            SetEditLayer(editLayer - 1);
            UpdateLayer();
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // 選択レイヤの表示優先度を下げるボタンを押した
        // -------------------------------------------------------------
        private void btnLayerDown_Click(object sender, EventArgs e)
        {
            SetUndo();

            if (lsCharas.Count <= 0 || editLayer >= lsCharas.Count - 1) return;
            Layer tmp = new Layer(lsCharas[editLayer]);
            lsCharas[editLayer] = lsCharas[editLayer + 1];
            lsCharas[editLayer + 1] = tmp;
            SetEditLayer(editLayer + 1);
            UpdateLayer();
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // レイヤの削除ボタンを押した
        // -------------------------------------------------------------
        private void btnLayerDelete_Click(object sender, EventArgs e)
        {
            if (lsCharas.Count <= 0) return;

            SetUndo();
            for (int i = editLayer; i < lsCharas.Count - 1; ++i)
            {
                lsCharas[i] = new Layer(lsCharas[i + 1]);
            }
            lsCharas.RemoveAt(lsCharas.Count - 1);
            while (editLayer >= lsCharas.Count) --editLayer;
            SetEditLayer(editLayer);
            UpdateLayer();
            picScreen.Visible = lsCharas.Count > 0;
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // カーソル位置の文字を任意の色に置換する
        // -------------------------------------------------------------
        private void RepTableColor(ExColor color)
        {
            if (DialogResult.No == CheckShowLayer()) return;
            lvLayer.Items[editLayer].Checked = true;

            Size sz = GetChrSize();
            if (rcSelect.IsEmpty)
            {
                // カーソル位置の色置換
                Point pnt = GetMapPos(picScreen.PointToClient(Cursor.Position));
                if (!(pnt.X >= 0 && pnt.X < sz.Width && pnt.Y >= 0 && pnt.Y < sz.Height)) return;

                SetUndo();
                lsCharas[editLayer].chrTable[pnt.X, pnt.Y].color = color;
            }
            else
            {
                // 指定範囲内の色置換
                SetUndo();
                for (int y = 0; y < rcSelect.Height; ++y)
                {
                    int posY = rcSelect.Y + y;
                    if (posY < 0 || posY >= sz.Height) continue;
                    for (int x = 0; x < rcSelect.Width; ++x)
                    {
                        int posX = rcSelect.X + x;
                        if (posX < 0 || posX >= sz.Width) continue;
                        lsCharas[editLayer].chrTable[posX, posY].color = color;
                    }
                }
            }
            CreateLayerPicture(lsCharas[editLayer]);
            lvLayer.Refresh();
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // 編集対象レイヤを変更する
        // -------------------------------------------------------------
        private void SetEditLayer(int n)
        {
            editLayer = n;
            if (lvLayer.SelectedItems.Count <= 0 || editLayer < lvLayer.Items.Count - 1) return;
            if (lvLayer.SelectedIndices[0] != editLayer) lvLayer.Items[editLayer].Selected = true;
            lvLayer.EnsureVisible(editLayer);
            lvLayer.Refresh();
        }

        // -------------------------------------------------------------
        // レイヤーリストを自分で描画する
        // -------------------------------------------------------------
        private void lvLayer_DrawItem(object sender, DrawListViewItemEventArgs e)
        {
            Rectangle rc = e.Bounds;
            rc.Y = rc.X = 0;
            int centerY = (rc.Height - 16) / 2;
            using (Bitmap bmp = new Bitmap(rc.Width, rc.Height))
            using (Graphics g = Graphics.FromImage(bmp))
            {
                if (editLayer == e.ItemIndex)
                {
                    g.FillRectangle(brLVItemSelect, rc);
                    g.DrawImage(layerThumbBack, 30, 6);
                    g.DrawRectangle(
                        pen: penGridLine,
                        x: 29,
                        y: 5,
                        width: layerThumbBack.Width + 2,
                        height: layerThumbBack.Height + 2
                    );
                    g.DrawRectangle(
                        pen: penGridSelect,
                        x: 27,
                        y: 3,
                        width: layerThumbBack.Width + 6,
                        height: layerThumbBack.Height + 6
                    );
                }
                else
                {
                    g.FillRectangle(brLVItem, rc);
                    g.DrawImage(layerThumbBack, 30, 6);
                    g.DrawRectangle(
                        pen: penGridLine,
                        x: 29,
                        y: 5,
                        width: layerThumbBack.Width + 2,
                        height: layerThumbBack.Height + 2
                    );
                }
                g.DrawLine(penGridLine, 0, rc.Height - 1, rc.Width, rc.Height - 1);
                g.DrawLine(penGridLine, 23, 0, 23, rc.Height);
                g.DrawLine(penGridLine, e.Bounds.Width -1, 0, e.Bounds.Width - 1, rc.Height);
                if (lsCharas[e.ItemIndex].thumbnail != null) g.DrawImage(lsCharas[e.ItemIndex].thumbnail, 30, 6);
                if (e.Item.Checked)
                {
                    g.DrawImage(checkedEye, 2, centerY, 16, 16);
                }
                else
                {
                    ControlPaint.DrawCheckBox(g, 2, centerY, 16, 16, ButtonState.Normal);
                }
                g.DrawString(
                    s: lvLayer.Items[e.ItemIndex].Text,
                    font: SystemFonts.MessageBoxFont,
                    brush: Brushes.White,
                    x: 16 + THUMBNAIL_WIDTH + 16 + 4,
                    y: centerY
                );
                e.Graphics.DrawImage(bmp, e.Bounds.X, e.Bounds.Y);
            }
        }

        // -------------------------------------------------------------
        // レイヤーアイテムが変更された
        // -------------------------------------------------------------
        private void lvLayer_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (reEnterListView) return;
            reEnterListView = true;
            if (lvLayer.SelectedItems.Count > 0)
            {
                SetEditLayer(lvLayer.SelectedItems[0].Index);
            }
            reEnterListView = false;
        }

        // -------------------------------------------------------------
        // レイヤーのチェックが変更された
        // -------------------------------------------------------------
        private void lvLayer_ItemChecked(object sender, ItemCheckedEventArgs e)
        {
            int idx = -1;
            if (!int.TryParse(e.Item.Name, out idx)) return;

            SetUndo();
            lsCharas[idx].isShow = e.Item.Checked;
        }

        // -------------------------------------------------------------
        // レイヤーのラベル編集が開始される
        // -------------------------------------------------------------
        private void lvLayer_BeforeLabelEdit(object sender, LabelEditEventArgs e)
        {
            editLabel = true;
        }

        // -------------------------------------------------------------
        // レイヤーのラベルが変更された
        // -------------------------------------------------------------
        private void lvLayer_AfterLabelEdit(object sender, LabelEditEventArgs e)
        {
            SetUndo();
            lsCharas[e.Item].label = e.Label;
            editLabel = false;
        }

        // -------------------------------------------------------------
        // レイヤーポップアップメニューが開かれる
        // -------------------------------------------------------------
        private void ctxLayerMenu_Opening(object sender, System.ComponentModel.CancelEventArgs e)
        {
            LayerJoinUnderToolStripMenuItem.Enabled = editLayer < lsCharas.Count - 1;
            bool isShow = false;
            foreach (Layer n in lsCharas)
            {
                isShow |= n.isShow;
            }
            LayerJoinLookToolStripMenuItem.Enabled = lsCharas.Count >= 2 && isShow;
            LayerJoinAllToolStripMenuItem.Enabled = lsCharas.Count >= 2;
            LayerRenameToolStripMenuItem.Enabled
                = LayerDupToolStripMenuItem.Enabled
                = LayerRemoveToolStripMenuItem.Enabled = lsCharas.Count > 0;
        }

        // -------------------------------------------------------------
        // メニュー／レイヤー名を変更する
        // -------------------------------------------------------------
        private void LayerRenameToolStripMenuItem_Click(object sender, EventArgs e)
        {
            lvLayer.Items[editLayer].BeginEdit();
        }

        // -------------------------------------------------------------
        // メニュー／レイヤーを追加する
        // -------------------------------------------------------------
        private void LayerAddToolStripMenuItem_Click(object sender, EventArgs e)
        {
            btnLayerAdd_Click(sender, e);
        }

        // -------------------------------------------------------------
        // メニュー／レイヤーを複製する
        // -------------------------------------------------------------
        private void LayerDupToolStripMenuItem_Click(object sender, EventArgs e)
        {
            lsCharas.Insert(editLayer, new Layer(lsCharas[editLayer]));
            SetEditLayer(editLayer);
            CreateLayerPicture(lsCharas[editLayer]);
            UpdateLayer();
        }

        // -------------------------------------------------------------
        // メニュー／レイヤーを削除する
        // -------------------------------------------------------------
        private void LayerRemoveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            btnLayerDelete_Click(sender, e);
        }

        // -------------------------------------------------------------
        // メニュー／下のレイヤーと結合
        // -------------------------------------------------------------
        private void LayerJoinUnderToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Size sz = GetChrSize();
            for (int y = 0; y < sz.Height; ++y)
            {
                for (int x = 0; x < sz.Width; ++x)
                {
                    if (lsCharas[editLayer].chrTable[x, y].code != 0) continue;
                    lsCharas[editLayer].chrTable[x, y] = lsCharas[editLayer + 1].chrTable[x, y];
                }
            }

            lsCharas.RemoveAt(editLayer + 1);
            SetEditLayer(editLayer);
            CreateLayerPicture(lsCharas[editLayer]);
            UpdateLayer();
        }

        // -------------------------------------------------------------
        // メニュー／表示レイヤーを結合
        // -------------------------------------------------------------
        private void LayerJoinLookToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Layer tmpChara = new Layer(lsCharas[editLayer]);
            tmpChara.chrTable = GetJoinLayer();
            for (int i = lsCharas.Count - 1; i >= 0; --i)
            {
                if (lsCharas[i].isShow) lsCharas.RemoveAt(i);
            }
            if (editLayer >= lsCharas.Count) editLayer = lsCharas.Count - 1;
            lsCharas.Insert(editLayer, tmpChara);
            for (int i = 0; i < lsCharas.Count; ++i)
            {
                CreateLayerPicture(lsCharas[i]);
            }
            lsCharas[editLayer].isShow = true;
            UpdateLayer();
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // メニュー／画像を結合
        // -------------------------------------------------------------
        private void LayerJoinAllToolStripMenuItem_Click(object sender, EventArgs e)
        {
            bool isShow = true;
            foreach (Layer n in lsCharas)
            {
                isShow &= n.isShow;
            }
            if (!isShow)
            {
                if (DialogResult.OK != MessageBox.Show(
                    text: "表示されていないレイヤーを破棄しますか？",
                    caption: productTitle,
                    buttons: MessageBoxButtons.OKCancel,
                    icon: MessageBoxIcon.Exclamation)
                )
                {
                    return;
                }
            }

            lsCharas[0].chrTable = GetJoinLayer();
            if (lsCharas.Count > 1)
            {
                lsCharas.RemoveRange(1, lsCharas.Count - 1);
            }
            editLayer = 0;
            CreateLayerPicture(lsCharas[0]);
            UpdateLayer();
            picScreen.Invalidate();
        }

        // -------------------------------------------------------------
        // 画像データがドロップされた
        // -------------------------------------------------------------
        private void MainForm_DragDrop(object sender, DragEventArgs e)
        {
            // ドロップされたデータがファイルパスか調べる
            if (!e.Data.GetDataPresent(DataFormats.FileDrop)) return;

            // ファイルを読み込む
            string[] paths = (string[])e.Data.GetData(DataFormats.FileDrop, false);
            for (int i = 0; i < paths.Count(); ++i)
            {
                if (i == 0 && (ModifierKeys & Keys.Control) != Keys.Control)
                {
                    FileOpen(paths[i]);
                }
                else
                {
                    FileMerge(paths[i]);
                }
            }
        }

        // -------------------------------------------------------------
        // 画像データがドラッグ中
        // -------------------------------------------------------------
        private void MainForm_DragEnter(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(DataFormats.FileDrop))
            {
                e.Effect = DragDropEffects.Copy;
            }
        }

        // -------------------------------------------------------------
        // コマンドラベルがクリックされた
        // -------------------------------------------------------------
        private void LabelCommand_Click(object sender, EventArgs e)
        {
            Point p = new Point(0, 0);
            ctxCmdMenu.Show(picCommand, p);
        }

        // -------------------------------------------------------------
        // フォームでキーが押された
        // -------------------------------------------------------------
        private void MainForm_KeyUp(object sender, KeyEventArgs e)
        {
            Point bakPos = posCursor;
            switch (e.KeyCode)
            {
                case Keys.Left:
                    --posCursor.X;
                    goto CHECK;
                case Keys.Right:
                    ++posCursor.X;
                    goto CHECK;
                case Keys.Up:
                    --posCursor.Y;
                    goto CHECK;
                case Keys.Down:
                    ++posCursor.Y;
CHECK:              if (!IsCurosrSelectRange())
                    {
                        posCursor = bakPos;
                    }
                    System.Windows.Input.KeyStates state = System.Windows.Input.Keyboard.GetKeyStates(System.Windows.Input.Key.Z);
                    if ((state & System.Windows.Input.KeyStates.Down) == System.Windows.Input.KeyStates.Down) OnClickPen();
                    break;
                case Keys.Back:
                    if (!textBox.Visible) return;
                    --posCursor.X;
                    if (IsCurosrSelectRange())
                    {
                        SetCharaTable(editLayer, new Chara(), posCursor);
                    }
                    else
                    {
                        posCursor = bakPos;
                    }
                    textBox.Text = "";
                    picScreen.Invalidate();
                    break;
                case Keys.Enter:
                    if (!textBox.Visible) return;
                    textBox.Visible = false;
                    SetCommand(Cmd.Pen);
                    picScreen.Focus();
                    break;
            }
        }

        // -------------------------------------------------------------
        // テキストボックスの入力内容が変化した
        // -------------------------------------------------------------
        private void textBox_TextChanged(object sender, EventArgs e)
        {
            if (reEdnterTextBox) return;
            reEdnterTextBox = true;

            // 選択範囲が無いまたは選択範囲内なのでキャラ文字を置く
            if (IsCurosrSelectRange() && textBox.Text.Length > 0)
            {
                Chara chr = new Chara(textBox.Text[0], charaSelect.color);
                SetCharaTable(editLayer, chr, posCursor);

                Point bakPos = posCursor;
                ++posCursor.X;
                if (!IsCurosrSelectRange())
                {
                    bakPos = posCursor;
                }
                textBox.Text = "";
                picScreen.Invalidate();
            }

            reEdnterTextBox = false;
        }
    }
}
