﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace CharaEdit40
{
    class M88t88
    {
        private static BinaryWriter _bw;    // バイナリライター
        private static int _elapse;         // 経過時間
        private byte[] _datas;              // 出力データ

        public M88t88(byte[] datas, UInt16 start, string gameFile) // gameFileは6文字まで
        {
            using (MemoryStream ms = new MemoryStream())
            using (_bw = new BinaryWriter(ms))
            {
                _elapse = 0;

                _bw.Write("PC-8801 Tape Image(T88)\0".ToCharArray());    // ファイルヘッダ
                write_ver();
                write_space(7928);
                write_mark(1322);
                write_filename(gameFile);
                write_mark(330);
                write_data(start, datas);
                write_mark(3966);
                write_space(9252);
                write_tag(0x0000, 0);   // 終了タグ

                _datas = ms.ToArray();
            }
        }
        public byte[] GetBinary() { return _datas; }

        private static void write_tag(int id, int len)
        {
            _bw.Write((ushort)id);
            _bw.Write((ushort)len);
        }

        private static void write_tick(int tick)
        {
            _bw.Write(_elapse);
            _bw.Write(tick);
            _elapse += tick;
        }

        // バージョン
        private static void write_ver()
        {
            write_tag(0x0001, 2);
            _bw.Write((ushort)0x0100);
        }

        // データタグ
        private static void write_data_tag(int datlen)
        {
            write_tag(0x0101, 12 + datlen);
            write_tick(44 * datlen);
            _bw.Write((ushort)datlen);
            _bw.Write((ushort)0x01cc);
        }

        // スペース(0)
        private static void write_space(int tick)
        {
            write_tag(0x0102, 8);
            write_tick(tick);
        }

        // マーク(1)
        private static void write_mark(int tick)
        {
            write_tag(0x0103, 8);
            write_tick(tick);
        }

        private static void write_filename(string filename)
        {
            string str = "$$$" + (filename + "      ").Remove(6);
            write_data_tag(9);
            _bw.Write(str.ToCharArray());
        }

        private static void write_data(int start, byte[] datas)
        {
            using (var fs = new MemoryStream(datas))
            {
                var buf = new byte[256];
                var blk = (int)Math.Ceiling(fs.Length / 128.0);
                write_data_tag((int)fs.Length + 7 + 3 * blk);

                // 開始アドレス
                buf[0] = (byte)(start >> 8);
                buf[1] = (byte)(start & 0xff);
                write_data_check(buf, 2);

                // データ
                while (true)
                {
                    var read = fs.Read(buf, 1, 128);
                    if (read <= 0) break;
                    buf[0] = (byte)read;
                    write_data_check(buf, 1 + read);
                }

                // 終了
                buf[0] = 0;
                write_data_check(buf, 1);
            }
        }

        private static void write_data_check(byte[] buf, int len)
        {
            int sum = 0;
            for (var i = 0; i < len; i++)
            {
                sum += buf[i];
            }
            int check = 0x100 - (sum & 0xff);   // チェックバイト
            _bw.Write((byte)0x3a);
            _bw.Write(buf, 0, len);
            _bw.Write((byte)check);
        }
    }
}
